% ----------------------------------------------------------------------
% System:	ECLiPSe Constraint Logic Programming System
% Copyright (C) Imperial College London and ICL 1995-1999
% Version:	$Id: alldiff_strong.pl,v 1.6 1999/08/07 15:49:58 js10 Exp $
% ----------------------------------------------------------------------

% ECLiPSe PROLOG LIBRARY MODULE
%
% $Id: alldiff_strong.pl,v 1.6 1999/08/07 15:49:58 js10 Exp $
%
% AUTHOR:
%	Stefano Novello, IC-Parc
%
% DESCRIPTION:
%
%	This is a strong version of alldifferent/1.
%
%	It looks at all pairs of minimum and maximum values in
% 	the domains of it's variables, for each pair (min,max)
%	it finds the number of variables whose value necessary lies
%	between min and max, and the union of their domains.
%	if the number of variables equals the size of the union
%	no other variable may take a value within that domain.
%	If the number of variable is greater than the union the
%	constrint fails.
%
%	The not_among constrint is left over from a previous
%	weaker implementation.
%	The algorithm is based roughly on:
%	Jean Francois Puget
%	A fast algorithm for the bound consistency of the alldiff constraint
%	in AAAI 1998.

:- module_interface(alldiff_strong).

:- lib(fd).
:- export
	alldifferent/1,		% alldifferent(List)
	not_among/2.		% not_among(X,List)


:- begin_module(alldiff_strong).
%
% not_among(X,ListOfVars)
%
not_among(X,ListWithoutX):-
	nonvar(X),
	call_priority(remove_element(ListWithoutX,X), 2).
not_among(X,ListWithoutX):-
	var(X),
	dvar_domain(X,DX),
	dom_size(DX,Size),
	count_subsets(DX,ListWithoutX,Size,1,Count,Others),
	make_suspension(not_among(X,ListWithoutX),4,Susp),
	insert_suspension(X,Susp,any of fd,fd),
	( Count == Size ->
	    call_priority(update_domains(Others), 2)
	;
	    true
	).

count_subsets(_,[],_,N,N,[]).
count_subsets(Dom,[H|T],Dom_Size,Sofar,N,Others):-
	dvar_domain(H,DH),
	(dom_difference(DH,Dom,Diff,_)->
            % DH not a subset of Dom
	    Others = [H-Diff|Others0],
	    count_subsets(Dom,T,Dom_Size,Sofar,N,Others0)
	;
            % DH subset of Dom
	    Count is Sofar + 1,
	    Count =< Dom_Size,
	    count_subsets(Dom,T,Dom_Size,Count,N,Others)
	).

update_domains([]).
update_domains([X-D|XDs]) :-
	dvar_update(X,D),
	update_domains(XDs).

remove_element([],_).
remove_element([X|Xs],E) :-
	dvar_remove_element(X,E),
	remove_element(Xs,E).

%
% alldifferent(ListOfVars)
%
alldifferent([]) :- !.
alldifferent([_]) :- !.
alldifferent(Vars) :-
	( foreach(t(Min,Max,Var),UnsortedTs),
	  foreach(Max,UnsortedMaxs),
	  foreach(Var,Vars)
	do
	    dvar_range(Var,Min,Max)
	),
	% all variables sorted in order of increasing Max
	TsByMax is sort(2,=<,UnsortedTs),
	% all max domains
	Maxs is sort(0,<,UnsortedMaxs),

	( fromto(Maxs,[Max|Maxs0],Maxs1,[]),
	  fromto([],TsLeMax0,TsLeMax,RemainingTs),			% remainingTs all vars s.t. maxdomain(var) <= largest(max)
	  fromto([],Mins0,Mins,_),
	  fromto(TsByMax,TsLeft0,TsLeft,TsLeftOver),			% Ts left Over all vars s.t maxdomain(var) >= largest max)
	  param(Propagated)
	do
	    le_than_max(Max,TsLeft0,UnsortedNewTs,TsLeft),
	    NewTs is sort(1,=<,UnsortedNewTs),
	    TsLeMax1 is merge(1,=<,TsLeMax0,NewTs),
	    get_minima(NewTs,UnsortedNewMins),
	    NewMins is sort(0,<,UnsortedNewMins),
	    Mins is merge(0,<,Mins0,NewMins),
	    % get_minima(NewTs,Mins),					% weaker propagation possible often faster.
	    ( fromto(Mins,[Min|M0],M,[]),				% iterate through all minima until a reducible range is found
	      fromto(TsLeMax1,TsInMinToMax0,TsInMinToMax,_),		% maintain list of Ts with domain that is a subset of [Min..Max]
	      fromto(TsLeMax1,T0,T,TsLeMax),				% maintain list of Ts < Max that are not removed from this alldiff
	      param(TsLeMax1,TsLeft,Max,Propagated)
	    do
	    	ge_than_min(Min,TsInMinToMax0,TsInMinToMax),
		% this unnecessary length means complexity is cubic.
		length(TsInMinToMax,SetSize),
		get_union_domain_size(TsInMinToMax,SetSize,UnionDom,RangeSize),
									% fail if too many in set
									% if we have found a range with N values and there
									% are N vars to occupy it
		( RangeSize == SetSize ->					% and our set of N vars is a strict subset of the vars
		    lt_min(Min,TsLeMax1,TsLtMin),
		    ( [TsLeft|TsLtMin] == [[]|[]]  ->
			M=M0,
			T=T0
		    ;
			sepia_kernel:get_priority(Pri),
			sepia_kernel:set_priority(2),
			remove_domain(UnionDom,TsLtMin),
			remove_domain(UnionDom,TsLeft),
			sepia_kernel:set_priority(Pri),
			wake,
			ts_to_vars(TsInMinToMax,VsInMinToMax),
			alldifferent(VsInMinToMax),
			Propagated = 1,
			M=[],
			T=TsLtMin
		    )
		;
		    M=M0,
		    T=T0
		)
	    ),
	    ( Propagated == 1 ->
	    	Maxs1=[]
	    ;
	    	Maxs1=Maxs0
	    )
	),
	( Propagated == 1 ->
		append(RemainingTs,TsLeftOver,NewTs),
		ts_to_vars(NewTs,NewVars),
		alldifferent(NewVars)
	;
		suspend(alldifferent(Vars),4,Vars->any)
	).
	    

	remove_domain(_Dom,[]).
	remove_domain(Dom,[t(_,_,Var)|Ts]) :-
	      dvar_domain(Var,VarDom),
	      dom_difference(VarDom,Dom,NewDom,_),
	      dvar_update(Var,NewDom),
	      remove_domain(Dom,Ts).

	ge_than_min(_Min,[],[]).
	ge_than_min(Min,[T|Ts],TsInMinToMax) :-
		T=t(MinT,_,_),
		( MinT < Min ->
		    ge_than_min(Min,Ts,TsInMinToMax)
		;
		    TsInMinToMax = [T|Ts]
		).

	lt_min(_Min,[],[]).
	lt_min(Min,[T|Ts],TsLtMin) :-
		T=t(MinT,_,_),
		( MinT < Min ->
		    TsLtMin = [T|TsLtMin0],
		    lt_min(Min,Ts,TsLtMin0)
		;
		    TsLtMin = []
		).

	le_than_max(_Max,[],[],[]).
	le_than_max(Max,[T|Ts],TsLeThanMax,RestTs) :-
		T=t(_,MaxT,_),
		( Max < MaxT ->
		    TsLeThanMax = [],
		    RestTs = [T|Ts]
		;
		    TsLeThanMax = [T|MoreTsLeThanMax],
		    le_than_max(Max,Ts,MoreTsLeThanMax,RestTs)
		).

	% this gets the size of the union of the domains. It
	% stops if the union is large enough i.e. > MinSize
	% in which case the output UnionDomain is not valid
	% since it will not be used.
	% also the domain size is just the domain size so far.
	get_union_domain_size([T|Ts],MinSize,UnionDomain,DomainSize) :-
		T=t(_,_,V),
		dvar_domain(V,DomV),
		dom_size(DomV,Size),
		( Size =<  MinSize ->
		    get_union_domain_size(Ts,DomV,MinSize,Size,UnionDomain,DomainSize)
		;
		    Size = DomainSize
		).

	get_union_domain_size([],UnionDomain,MinSize,DomainSize,UnionDomain,DomainSize) :-
		DomainSize >= MinSize.
	get_union_domain_size([T|Ts],PreviousUnion,MinSize,_,UnionDomain,DomainSize) :-
		T=t(_,_,V),
		dvar_domain(V,DomV),
		dom_union(PreviousUnion,DomV,Union,Size),
		( Size =< MinSize ->
		    get_union_domain_size(Ts,Union,MinSize,Size,UnionDomain,DomainSize)
		;
		    Size = DomainSize
		).

	get_union_domain([V|Vs],UnionDomain) :-
		dvar_domain(V,FirstDom),
		( foreach(Var,Vs),
		  fromto(FirstDom,D0,D,UnionDomain)
		do
		    dvar_domain(Var,VarDom),
		    dom_union(D0,VarDom,D,_)
		).

	get_minima(Ts,Mins):-
		( foreach(t(Min,_,_),Ts),
		  foreach(Min,UnsortedMins)
		do
		    true
		),
		sort(0,<,UnsortedMins,Mins).

	ts_to_vars(Ts,Vars):-
		( foreach(t(_,_,Var),Ts),
		  foreach(Var,Vars)
		do
		    true
		).

t(1,[A,B,B,C,D,E,F,G,H,I]):-
	 A::2..8,
	 B::2..8,
	 C::4..10,
	 D::4..10,
	 E::2..8,
	 F::6..12,
	 G::4..10,
	 H::1..5,
	 I::7..13,
	 alldifferent([2,A,B,C,D,3,E,F,G,H,12,I]).



t(2,[A,B,C,D,E]) :-
	A::1..10,
	B::2..9,
	C::2..3,
	D::3..4,
	E::2..3,
	Vars = ([A,B,C,D,E]),
	alldifferent(Vars),
	printf("%mw%n%b",[Vars]).
t(3,[A,B,C,D,E]) :-
	A::1..10,
	B::2..9,
	C::3..4,
	D::3..4,
	E::2..3,
	Vars = ([A,B,C,D,E]),
	alldifferent(Vars),
	printf("%mw%n%b",[Vars]).
t(4,[A,B,C,D,E]) :-
	A::1..10,
	B::2..9,
	C::3..4,
	D::[2,4],
	E::2..3,
	Vars = ([A,B,C,D,E]),
	alldifferent(Vars),
	printf("%mw%n%b",[Vars]).
