% ----------------------------------------------------------------------
% System:	ECLiPSe Constraint Logic Programming System
% Copyright (C) Imperial College London and ICL 1995-1999
% Version:	$Id: calendar.pl,v 1.4 1999/08/07 15:50:02 js10 Exp $
% ----------------------------------------------------------------------

%
% ECLiPSe PROLOG LIBRARY MODULE
%
% $Id: calendar.pl,v 1.4 1999/08/07 15:50:02 js10 Exp $
%
% IDENTIFICATION:	calendar.pl
%
% AUTHOR:		Joachim Schimpf, IC-Parc
%
% CONTENTS: 		Some routines for calendar computations,
%			based on modified julian dates (MJD).
%
% DESCRIPTION:
%
%	Julian Dates (JD) and Modified Julian Dates (MJD) are a
%	consecutive day numbering scheme widely used in astronomy,
%	space travel etc. It is defined for a long period from
%	12 noon, 1 Jan 4713 BC  to  12 noon, 1 Jan 3268 AD.
%
%	Here we use MJDs as the central representation (JDs are a bit
%	awkward because they change at noon and are very large numbers).
%
%	Note that you can use fractional MJDs to denote the time of day.
%	The time is then defined to be Universal Time (UT, formerly GMT).
%	That means that every day has a unique integer number,
%	or every time point has a unique float representation!
%
%	Differences between times are obviously trivial to compute,
%	and so are weekdays (by simple mod(7) operation).
%
%	The following code is valid for dates between
%		 1 Jan 0001 = MJD -678577 = JD 1721423.5
%	and
%		22 Jan 3268 = MJD  514693 = JD 2914693.5
%
%	The relationship between JD and MJD is simply
%	MJD = JD-2400000.5, ie MJD 0 = 17 Nov 1858.
%
% Implementation note:
%	To simplify the leap year computations, we work internally with years
%	that start on the 1st of Mar with day 0 and end on 28 or 29 Feb of
%	the next year with day 364 or 365. Also, we internally use a day
%	numbering starting from (nonexistent) 1.3.0000.
%
%	The 15 Oct 1582 (MJD -100840, JD 2299160.5) marks the starting
%	date of the gregorian calendar reform (modified leap year rule).
%---------------------------------------------------------------------

:- module_interface(calendar).

:- export
	date_to_mjd/2,		% date_to_mjd(+D/M/Y, -MJD)
	mjd_to_date/2,		% mjd_to_date(+MJD, -D/M/Y)
	mjd_to_weekday/2,	% mjd_to_weekday(+MJD, -DayName)
	mjd_to_dow/3,		% mjd_to_dow(+MJD, +FirstWeekday, -DoW)
	mjd_to_dow/2,		% mjd_to_dow(+MJD, -DoW)
	mjd_to_time/2,		% mjd_to_time(+MJD, -H:M:S)  GMT/UT
	time_to_mjd/2,		% time_to_mjd(+H:M:S, -MJD)  MJD < 1.0
	mjd_to_dy/2,		% mjd_to_dy(+MJD, -DoY/Y)
	dy_to_mjd/2,		% dy_to_mjd(+DoY/Y, -MJD)
	mjd_to_dwy/2,		% mjd_to_dwy(+MJD, -DoW/WoY/Y)
	mjd_to_dwy/3,		% mjd_to_dwy(+MJD, +FirstWeekday, -DoW/WoY/Y)
	dwy_to_mjd/2,		% dwy_to_mjd(+DoW/WoY/Y, -MJD)
	dwy_to_mjd/3,		% dwy_to_mjd(+DoW/WoY/Y, +FirstWeekday, -MJD)

	unix_to_mjd/2,		% unix_to_mjd(+UnixSec, -MJD)
	mjd_now/1,		% mjd_now(-MJD)

	jd_to_mjd/2,		% jd_to_mjd(+JD, -MJD)
	mjd_to_jd/2.		% mjd_to_jd(+MJD, -JD)


%---------------------------------------------------------------------
:- begin_module(calendar).
:- pragma(expand).
:- pragma(nodebug).
%---------------------------------------------------------------------

date_to_mjd(Din/Min/Yin, MJD) :-
	date_to_internal(Din, Min, Yin, DayNr, Y),
	dy_leap_days(DayNr, Y, Leapdays),
	MJD is DayNr + 365*Y + Leapdays - 678883.


mjd_to_date(MJD, Dout/Mout/Yout) :-
	mjd_to_internal(MJD, Days, Leapdays),
	MJD1 is MJD + 1,
	mjd_to_internal(MJD1, _Days1, Leapdays1),
	Y is (Days - Leapdays1)//365,
	DayNr is Days - Y*365 - Leapdays,
	date_from_internal(Dout, Mout, Yout, DayNr, Y).


:- mode mjd_to_internal(+,-,-).
mjd_to_internal(MJD, DD, Leapdays) :-
	MJD >= -100840, !, 		% after 15 Oct 1582: gregorian
	DD is fix(MJD + 678881),	% 1 Mar 0000 if always gregorian
	Mod400 is DD mod 146097,	% mod days in 400 gregorian years
	Mod100 is Mod400 mod 36524,	% mod days in 100 gregorian years
	Leapdays is (DD//146097)*97 + (Mod400//36524)*24 + (Mod100//1461).
mjd_to_internal(MJD, DD, Leapdays) :-
	DD is fix(MJD + 678883),	% 1 Mar 0000 for real
	Leapdays is DD//1461.		% julian rule


:- mode dy_leap_days(+,+,-).
dy_leap_days(DayNr, Y, Leapdays) :-
	Y*365 + DayNr >= 577658, !, 		% after 15 Oct 1582?
	Leapdays is Y//4 - Y//100 + Y//400 + 2.	% gregorian rule
dy_leap_days(_, Y, Leapdays) :-
	Leapdays is Y//4.			% julian rule


date_to_internal(D, M, Y, DayNr, Ynorm) :-
	month_offset(M, Moffs, Yoffs),
	DayNr is Moffs + D - 1,
	Ynorm is Y + Yoffs.

date_from_internal(D, M, Y, DayNr, Ynorm) :-
	month_offset(Month, Moffs, Yoffs),
	DayNr >= Moffs,
	!,
	M = Month,
	D is DayNr - Moffs + 1,
	Y is Ynorm - Yoffs.


% month_offset(Month, MonthOffs, YearOffs)
% month_offset(1..12,    0..365,    -1..0)
:- mode month_offset(?, -, -).	% clause order important!
month_offset( 2, 337, -1).	% Feb
month_offset( 1, 306, -1).	% Jan
month_offset(12, 275,  0).	% Dec
month_offset(11, 245,  0).	% Nov
month_offset(10, 214,  0).	% Oct
month_offset( 9, 184,  0).	% Sep
month_offset( 8, 153,  0).	% Aug
month_offset( 7, 122,  0).	% Jul
month_offset( 6,  92,  0).	% Jun
month_offset( 5,  61,  0).	% May
month_offset( 4,  31,  0).	% Apr
month_offset( 3,   0,  0).	% Mar


%---------------------------------------------------------------------

mjd_to_time(MJD, H:M:S) :-
	Frac is MJD-floor(MJD),
	H is fix(Frac*24),
	M is fix(Frac*1440) mod 60,
	S is Frac*86400 - H*3600 - M*60.

time_to_mjd(H:M:S, MJD) :- !,
	MJD is ((H*60 + M)*60 + S)/86400.
time_to_mjd(H:M, MJD) :-
	MJD is (H*60 + M)/1440.

%---------------------------------------------------------------------

unix_to_mjd(Seconds_since_1970, MJD) :-
	MJD is 40587 + Seconds_since_1970/86400.

mjd_now(MJD) :-
	get_flag(unix_time, T),
	unix_to_mjd(T, MJD).

%---------------------------------------------------------------------

jd_to_mjd(JD, MJD) :-
	MJD is JD-2400000.5.

mjd_to_jd(MJD, JD) :-
	JD is MJD+2400000.5.

%---------------------------------------------------------------------
% CAUTION: Make sure argument of fix() is always positive!!!

mjd_to_dy(MJD, DoY/Y) :-
	mjd_to_date(MJD, _/_/Y),
	DoY is fix(MJD - date_to_mjd(1/1/Y)) + 1.

dy_to_mjd(DoY/Y, MJD) :-
	MJD is date_to_mjd(1/1/Y) + DoY - 1.
	

mjd_to_dow(MJD, DoW) :-
	mjd_to_dow(MJD, monday, DoW).

mjd_to_dow(MJD, FirstWeekday, DoW) :-
	DoW is fix(MJD + 678881 + wd_nr(FirstWeekday)) mod 7 + 1.

    wd_nr(monday, 2).
    wd_nr(tuesday, 1).
    wd_nr(wednesday, 0).
    wd_nr(thursday, 6).
    wd_nr(friday, 5).
    wd_nr(saturday, 4).
    wd_nr(sunday, 3).


mjd_to_dwy(MJD, DWY) :-
	mjd_to_dwy(MJD, monday, DWY).

mjd_to_dwy(MJD, FirstWeekday, DoW/WoY/Y) :-
	mjd_to_date(MJD, _/_/Y),
	date_to_mjd(1/1/Y, Jan1st),
	mjd_to_dow(Jan1st, FirstWeekday, Jan1stW),
	mjd_to_dow(MJD, FirstWeekday, DoW),
	WoY is fix((MJD-Jan1st) - (DoW-Jan1stW) + 7) // 7.

dwy_to_mjd(DWY, MJD) :-
	dwy_to_mjd(DWY, monday, MJD).

dwy_to_mjd(DoW/WoY/Y, FirstWeekday, MJD) :-
	date_to_mjd(1/1/Y, Jan1st),
	mjd_to_dow(Jan1st, FirstWeekday, Jan1stW),
	MJD is Jan1st + 7*WoY + (DoW-Jan1stW) - 7.


mjd_to_weekday(MJD, WD) :-
	I is fix(MJD + 678881) mod 7 + 1,	% avoid negative modulus!
	arg(I, wd(wednesday, thursday, friday,
			saturday, sunday, monday, tuesday), WD).

/***
%-----------------------------------------------------------------------
% exhaustive test
% should only print the 10 missing days 5..14 Oct 1582
%-----------------------------------------------------------------------
test_all :-
	setval(mjd, -678577), between(1,    3267, 1, Y),
%	setval(mjd, -101117), between(1582, 3267, 1, Y),

	put(13), write(Y), flush(output),
	between(1, 12, 1, M),
	(leap(Y) ->
	    arg(M, d(31,29,31,30,31,30,31,31,30,31,30,31), MD)
	;
	    arg(M, d(31,28,31,30,31,30,31,31,30,31,30,31), MD)
	),
	between(1, MD, 1, D),
	DateIn = D/M/Y,
	date_to_mjd(DateIn, MJD),
	mjd_to_date(MJD, DateOut),
	( DateOut == DateIn ->
	    getval(mjd, MJDexp),
	    ( MJD =:= MJDexp ->
		incval(mjd)
	    ;
		printf("wrong MJD generated: %w\n", date_to_mjd(DateIn, MJDexp) -> MJD)
	    )
	;
	    printf("reverse conversion failed: %w\n", mjd_to_date(MJD, DateIn) -> DateOut)
	),
	fail.
test_all.

leap(Y) :-
	Y mod 4 =:= 0,
	( Y < 1582 ->
	    true
	; Y mod 100 =:= 0 ->
	    Y mod 400 =:= 0
	;
	    true
	).

now :-
	mjd_now(MJD),
	mjd_to_date(MJD, Date),
	mjd_to_time(MJD, Time),
	mjd_to_weekday(MJD, W),
	writeln((W,Date,Time,mjd=MJD)).
***/
