% ----------------------------------------------------------------------
% System:	ECLiPSe Constraint Logic Programming System
% Copyright (C) Imperial College London and ICL 1995-1999
% Version:	$Id: cprolog.pl,v 1.5 1999/08/07 15:50:04 js10 Exp $
% ----------------------------------------------------------------------

/*
 * SEPIA PROLOG SOURCE MODULE
 */

/*
 * sccsid("%W%		%E%").
 * sccscr("%Z%  Copyright 1989 ECRC GmbH ").
 */

/*
 * IDENTIFICATION:	cprolog.pl
 *
 * DESCRIPTION: 	C-Prolog compatibility package.
 *			This is also the basis for the
 *			Quintus and SICStus packages.
 *
 * CONTENTS:     
 *
 */

:- module_interface(cprolog).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
%	SYNTAX
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
:-
	set_flag(syntax_option, nl_in_quotes),
	set_flag(syntax_option, no_blanks),
        set_flag(syntax_option, no_array_subscripts),
	set_flag(syntax_option, limit_arg_precedence),

	set_chtab(0'$, lower_case),
	set_chtab(0'&, symbol),
	set_chtab(0'\, symbol),
	set_chtab(128, string_quote),	% there must be some string_quote
	set_chtab(0'", list_quote),

	op(0,   xfx, (of)),
	op(0,   xfx, (with)),
	op(0,   xfy, (do)),
	op(0,   xfx, (@)),
	op(0,   fx, (-?->)),
	define_macro((with)/2, (=)/2, [local]),
	define_macro((of)/2, (=)/2, [local]),

	op(500, fx, (\)),
	op(900, fy, (spy)),
	op(900, fy, (nospy)).

/*
 * INCLUDED GLOBAL LIBRARIES
 */

:- use_module(library(cio)).
:- use_module(library(scattered)).

:- begin_module(cprolog).

:- system.		% compiler directive to add the SYSTEM flag

% Make sure that the important operators are ok
:-	op(1100, xfy, ;),
	op(1050, xfy, ->).

/*
 * IMPORT DIRECTIVES
 */

:- import
	assert_/2,
	current_predicate_body/2,
        error_handler/2,
	get_flag_body/4,
        import_body/2,
	set_default_error_handler/2,
        system_error_handler/2,
	undef_dynamic_handler/3
   from sepia_kernel.

/*
 * OTHER DIRECTIVES
 */

:- local
	get/1,
	instance/2,
	put/1,
	current_predicate/2.		% because it will be declared skipped

:-  tool(assertz/1, assert_/2),
    tool(consult/1, compile/2),
    tool(reconsult/1, compile/2),
    tool(current_predicate/2, current_predicate_body/3).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
%	EVENT HANDLERS
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

cp_inst_fault(_, arg(N, _, _)) :- integer(N), !, fail.
cp_inst_fault(_, functor(_, _, _)) :- !, fail.
cp_inst_fault(X, Y):- error_handler(X,Y).

cp_range_error(_, arg(_, _, _)) :- !, fail.
cp_range_error(X, Y) :- error_handler(X, Y).

cp_type_error(_, arg(_, _, _)) :- !, fail.
cp_type_error(X, Y) :- error_handler(X, Y).

% This simulates Quintus' behaviour
cp_undef_dynamic_handler(_, retract_all(_), _) :- !.
cp_undef_dynamic_handler(_, listing(_), _) :- !.
cp_undef_dynamic_handler(_, retract(_), _) :- !, fail.
cp_undef_dynamic_handler(_, clause(_), _) :- !, fail.
cp_undef_dynamic_handler(_, clause(_, _), _) :- !, fail.
cp_undef_dynamic_handler(E, G, M) :-
	undef_dynamic_handler(E, G, M).

cp_access_undefined(_, abolish(_)) :-
	!.
cp_access_undefined(X,Y) :-
	error_handler(X,Y).

nofileerrors_handler(2, open(_, _, _)) :- !, fail.
nofileerrors_handler(2, Goal) :-
	system_error_handler(2, Goal).

fileerrors :-
	set_default_error_handler(170, system_error_handler/2),
	set_default_error_handler(171, error_handler/2),
	reset_error_handler(170),
	reset_error_handler(171).

nofileerrors :-
	set_default_error_handler(170, nofileerrors_handler/2),
	set_default_error_handler(171, fail/0),
	reset_error_handler(170),
	reset_error_handler(171).

:-
	% we may change the default handlers, because we can't switch back
	set_default_error_handler(4, cp_inst_fault/2),
	reset_error_handler(4),
	set_default_error_handler(5, cp_type_error/2),
	reset_error_handler(5),
	set_default_error_handler(6, cp_range_error/2),
	reset_error_handler(6),
	set_default_error_handler(60, cp_access_undefined/2),
	reset_error_handler(60),
	set_default_error_handler(70, cp_undef_dynamic_handler/3),
	reset_error_handler(70),
	set_default_error_handler(100, cp_undef_dynamic_handler/3),
	reset_error_handler(100).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
%	NEW PREDICATES
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

sh :-
	getenv('SHELL', X),
	sh(X).

primitive(X) :- atomic(X), fail_if(atom(X)).

db_reference(X) :- type_of(X, db_reference).

get0(X):- get(input, X).

leash(Ports) :-
	set_leash(_, print),
	set_leash(Ports, stop).


heapused(X) :- X is statistics(general_heap_used) + statistics(code_heap_used).

prompt(Old, New) :-
	get_prompt(input, Old, Out),
	set_prompt(input, New, Out).

ttyput(Char) :-
	N is Char,
	put(stdout, N).

current_functor(F, T) :-
	current_functor(F/A),
	functor(T, F, A).

erased(Ref) :-
	fail_if(referenced_record(Ref, _)).

%  ARITHMETIC

log10(Y,X):- X is ln(Y)/ln(10.0).

log(Y,X) :- X is ln(Y).

.(X,_,X).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
%	REDEFINED PREDICATES
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

get(X):- repeat, get(input, X), X > 32, X < 127, !.

put(X):- N is X, put(output,N).

instance(Ref, Term) :-
	referenced_record(Ref, Term).

current_predicate_body(F, T, M) :-
	current_predicate_body(F/A, M),
	functor(T, F, A),
	get_flag_body(F/A, definition_module, M, M).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
%	MODULE INITIALIZATION
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% none

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
%	EXPORTED AND SKIPPED PREDICATES
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

:- export
	(.)/3,		% to evaluate lists in arithmetic expressions
	assertz/1,
	consult/1,
	current_functor/2,
	current_predicate/2,
	db_reference/1,
	erased/1,
	fileerrors/0,
	get/1,
	get0/1,
	heapused/1,
	instance/2,
	leash/1,
	log10/2,
	log/2,
	nofileerrors/0,
	primitive/1,
	prompt/2,
	put/1,
	reconsult/1,
	sh/0.

:- skipped
	assertz/1,
	consult/1,
	current_functor/2,
	erased/1,
	fileerrors/0,  
	get/1, 
	get0/1, 
	instance/2,
	log10/2,
	nofileerrors/0,
	primitive/1,
	prompt/2,
	put/1,
	reconsult/1.
