% ----------------------------------------------------------------------
% System:	ECLiPSe Constraint Logic Programming System
% Copyright (C) Imperial College London and ICL 1995-1998
% Copyright (C) Parc Technologies Ltd 1999
% Version:	$Id: dichotomic.pl,v 1.2 1999/08/07 14:56:36 js10 Exp $
% ----------------------------------------------------------------------

% Emacs      : --*-prolog-*--
% File       : dichotomic.pl
% Author     : Vassilis Liatsos <vl@doc.ic.ac.uk>
% Copyright  : ic-parc, imperial college
% Description: an implementation of the dichotomic search for eclipse
% Date       : June 1998

% Usage
% 
% Load module using: use_module(dichotomic).
% Module provides two predicates: 
%  dichotomic/3 given a goal and a cost term
%  dichotomic/5 same as above with template (see min_max)

:- module_interface(dichotomic).

:- lib(fd).
:- export 
      dichotomic/3, 
      dichotomic/5.


:- tool(dichotomic/3), tool(dichotomic/5).




:- begin_module(dichotomic).

:- make_local_array(sol_up),
   make_local_array(solution).

:- define_error("Dichotomic factor not within allowed range 0..1",N), 
   setval(di_usererror,N).


% dichotomic(+Goal,?Cost,Factor,Module)
:- tool(dichotomic/3,dichotomic/4).
dichotomic(Goal,Cost,Factor,Module):-
	dichotomic(Goal,Goal,Goal,Cost,Factor,Module).

% dichotomic(+Goal,?Template,?Solution,?Cost,Factor,Module)
:- tool(dichotomic/5,dichotomic/6).
dichotomic(Goal,Template,Solution,Cost,Factor,Module):-
	( not nonground(Factor), 0=< Factor, Factor =< 1 ->
	    true
	;   getval(di_usererror,N),
            error(N,dichotomic) ),
	setval(sol_up,no),
	dvar_range(Cost,Min,_CMax),
	register_initial_solution(Goal,Template,Cost,Max,Module),
	NewMax is Max-1,
	setval(sol_up,no),
	search(Goal,Template,Cost,Min,NewMax,Module,Factor),
	getval(solution,sol(Solution,Cost)).

register_initial_solution(Goal,Template,Cost,_,Module):-
	once(call(Goal,Module)),
	% register solution and fail
	register_solution(Template,Cost).
register_initial_solution(Goal,_Template,_Cost,Max,_):-
	% an initial solution was produced (otherwise no solution exists)
	(getval(sol_up,yes) ->
	    getval(solution,sol(_,Max))
	;
            printf('no feasible solution for %w\n',[Goal]),fail
	).

% register_solution(+Template,+Cost)
% given a ground goal and a ground cost registers solution and fails.
register_solution(Template,Cost):-
	setval(sol_up,yes),
	printf('Found a solution of cost %w\n%b',[Cost]),
	setval(solution,sol(Template,Cost)),!,fail.


search(_Goal,_Template,_Cost,Lb,Ub, _, _):-
	% Terminating condition
	Lb > Ub,!.

search(Goal,Template,Cost,Lb,Ub,Module,Factor):-
	Lb =< Ub,
	setval(sol_up,no),
	Temp is fix(Lb + (Ub-Lb)*Factor),
	Cost #<= Temp,
	Cost #>= Lb, 
	printf('Searching for solution in cost range [%w..%w]\n%b',[Lb,Temp]),
        once(call(Goal,Module)),
	register_solution(Template,Cost).

search(Goal,Template,Cost,Lb,Ub,Module,Factor):-
	Lb =< Ub,
	(getval(sol_up,no)->
	    % Solution not updated therefore new lower bound
	    NewLb is fix(Lb + (Ub-Lb)*Factor) + 1, 
	    Tmp is NewLb -1,
	    printf('No solution in cost range  [%w..%w]\n%b',[Lb,Tmp]),
	    search(Goal,Template,Cost,NewLb,Ub,Module,Factor)
	;
	    % Solution updated therefore new upper bound
	    getval(solution,sol(_,NewUb)),
	    NewUb1 is NewUb-1,
	    search(Goal,Template,Cost,Lb,NewUb1,Module,Factor)
	).


end_of_file.
