% ----------------------------------------------------------------------
% System:	ECLiPSe Constraint Logic Programming System
% Copyright (C) Imperial College London and ICL 1995-1998
% Copyright (C) Parc Technologies Ltd 1999
% Version:	$Id: ech2pl.pl,v 1.5.2.1 1999/08/20 17:50:19 ks15 Exp $
% ----------------------------------------------------------------------

chrcompile(Error, Culprit, Module) :- Culprit = (term,_,_), !,
   error(default(Error), Culprit, Module).
chrcompile(Error, Culprit, Module) :- Culprit = (_, _, dumped), !,
   error(default(Error), Culprit, Module).
chrcompile(Error, Culprit, Module) :-
   recorded_list('CHRadding_code', Modules),
   erase_all('CHRadding_code'),
   (Modules \== [] -> % have read in some constraints 
        chrcompile_by_module(Modules) ; true
   ),
   error(default(Error), Culprit, Module).

chrcompile_by_module([]) :- !.
chrcompile_by_module([Module|Modules]) :-
   recorded('CHRconst_count', count(Module,N)),
   % N should be > 0
   recorded_list('CHRconstraints', Constraints)@Module,
   recorded_list('CHRcode', Rules)@Module,
   compile_term([(:- import sepia_kernel), (:- set_flag(coroutine, on)),
          number_of_constraints(N)])@Module,
   transform(Rules,Constraints,N, Module),
   ArraySize is N + 1, % need +1 to avoid using 0.
   (current_array('CHRcdelete_count'(OldSize),_)@Module ->
      (OldSize \== ArraySize ->
         erase_array('CHRcdelete_count'/1)@Module,
         make_local_array('CHRcdelete_count'(ArraySize), integer)@Module
       ; true
      )
     ;make_local_array('CHRcdelete_count'(ArraySize), integer)@Module

   ), 
   chrcompile_by_module(Modules).

check_if_new_or_update(Module) :-
   (recorded('CHRadding_code', Module) ->
        true  
           
      ; % indicates that CHR code is being added since last (if any) compile
        recorda('CHRadding_code', Module)

   ), 
   (recorded('CHRconst_count', count(Module,_)) ->
    % CHRconst_count will be defined for Module if CHR encountered for Module
       true ; initialise_module_for_chr(Module)
   ).

:- set_event_handler(139, chrcompile/3).

/* transform(+Rules, +Constraints, +N, +Module)
   transforms the list of CHR in rules into Prolog code. Constraints is the 
   list of declared constraints for these rules. N is the number of
   constraints, and Nprule is the number of propagation rules in the CHR
*/ 
transform(Rules, Constraints, N, Module) :-
    initialise_processed(N, ProcessedRules),
    syntax_check(Rules, Constraints, ProcessedRules, Module),
    recorded('CHRprule_count', count(Module,Nprule)),
    compile_term(number_of_propagations(Nprule))@Module,
    translate(ProcessedRules, Constraints, Nprule, Module).


indexing_argsize(_N, 0).% :- writeln(_N). % for now

  
/* translate(+Processed, +Constraints, +Nprule, +Module)
   translate the processed rules in Processed into Prolog code. Constraints is
   the list of constraints, and Nprule is the number of propagation rules --
   if this is zero, the suspension can be simplified.
*/
translate(Processed,Constraints,Nprule, Module) :- 
    functor(Processed,_,Size),
    translate_each_constraint(Size, Processed, Constraints, Nprule, Module).

/* translate_each_constraint(+Nth, +Processed, +Constraints, +Nprule, +Module)
   translate the rules associated with the Nth constraint, i.e. those rules
   where the Nth constraint occurs in the head. Note that seperate code needs to 
   be generated for each occurance of a head constraint, except where the
   situation is symmetric and can be optimised
*/
translate_each_constraint(0,_,_,_,_) :- !.
translate_each_constraint(N, Processed, Constraints, Nprule, Module) :-
   arg(N,Processed,HRuleList),
   length(HRuleList, NRule),
   indexing_argsize(NRule, NRule1),
   nth_member(N, Constraints, F/A),
   printf(user, "Found CHR %a/%w, generating transformed code\n", [F,A]),
   gen_code_for_constraint(HRuleList, [], F, A, 1, NRule1, Nprule, Code0, Module),
   optimise(Code0, Code),
%   printcode(Code), 
   call(compile_term([(:- pragma(expand))|Code]), Module),
   N1 is N - 1,
   translate_each_constraint(N1, Processed, Constraints, Nprule, Module).

/* construct_exec_constraint(+WrappedF, +WrappedA, +F, +A, +ConstNum, +NRule, +Nprule, -Code, +Module)
    constructs the various codes for managing a particular constraint. Two 
    clauses are needed: one (WrappedF/WrappedA) that is suspended, and when 
    woken, calls the rules (with possible optimisations); and the other, 
    called from the plane constraint (F/A), which is used to initiate a
    new call to the rules. A demon directive for WrappedF/WrappedA is also
    needed so that it is not killed when woken. ConstNum is the index number
    into the constraint store for this constraint (F/A)
*/
construct_exec_constraint(WF, WA, CF, CA, ConstNo, NRule, Nprule, Code, Module) :-
    functor(WHead, WF, WA), % wrapped constraint 
    make_current_rule_name(CF, CA, 1, NRule, Nprule, FirstRule0, A),
    WHead =.. [_F,_ConsNum,Constraint|MetaArgs], 
    % WHead - suspended constraint, for wake up calls
    functor(Constraint,CF,CA),
    Constraint =.. [CF|ConsArgs], % original constraint form
    length(Index,NRule),
    instantiate_list(Index, 1),  % make all 1 for now 
    append(Index, ConsArgs, L1),
    append(L1, MetaArgs, ArgList),
    functor(FirstRule0, FirstRuleF, A),
    Body =.. [FirstRuleF|ArgList],
    add_check_in_store(Constraint, Body, ConstNo, Call, Module),
    % to add optimising code before Body in WHead
    Code = [(:- demon(WF/WA)), ( WHead :- Body), (Constraint :- Call)].


add_check_in_store(Constraint, Call, No, Check, Module) :-
   recorded('CHRdont_in_store', Module) ->
         Check = Call
       ; Check = (call_explicit(get_global_constraint(No, SuspL), ech),
                    (call_explicit(is_in_store(SuspL, Constraint), ech) -> true ; Call)
                 ).


/* extract the first Ath argument from the structure NewHead as a list, ConsArgs */
get_constraint_args(NewHead, A, ConsArgs) :-
 NewHead =.. [_|Args],
 split_list(A, Args, ConsArgs, _).


gen_code_for_constraint([], PreviousTryNexts, F, A, Nth, NRule, Nprule, CodeTail, Module) :- !,
  % need to generate termination rule
  (PreviousTryNexts \== none -> % have processed rules for this constraint
    make_rulefunctor(F, A, Nth, Functor),
    nmetaargs(Nprule, N),
    NewArity is A + N, % no indexing for this last clause.
    functor(NewHead, Functor, NewArity),
    get_constraint_args(NewHead, A, ConsArgs),
    Head =.. [F|ConsArgs], % set the constraint arguments
    AddBase is A + 1, % start of extra args
    arg(AddBase, NewHead, Index),
    %%% MAGIC NUMBER
    InPos is AddBase + 1,
    arg(InPos, NewHead, In),
    (Nprule =:= 0 -> ExtraArgs = [Index,In] 
          ; AppPos is AddBase + 2, arg(AppPos, NewHead, Applied),
            ExtraArgs = [Index,In,Applied]
    ),

    get_constraintnumber_fa(F,A, ConstNo, Module),
    construct_wrapper_constraint(Head, ExtraArgs, Constraint, _),
    construct_writeconstraint_code(Constraint, WriteDefine, WriteRule),
    CodeTail = [( NewHead :-
      (var(Index) -> 
          call_explicit(suspend_constraint(Constraint, Head, ConstNo),ech)
        ; suspension_to_goal(Index, Goal, _),
          setarg(wrapper_inpos, Goal, In)
      )
    ),WriteDefine, WriteRule| ExecConstraint],
    functor(Constraint, WH, WA),

    construct_exec_constraint(WH,WA,F,A, ConstNo, NRule, Nprule, 
         ExecConstraint, Module),
    construct_previous_try_next(PreviousTryNexts, A, NRule, NRule, NewHead)
  ; /* no code for this constraint at all */
    printf(user, "Warning: No rule found for constraint %w/%w.\n",[F,A]),
    CodeTail = []).
gen_code_for_constraint([rule(Rule,Status,Head,KeepHeads,DeleteHeads,Body,
    PInfo, ComInfo, _Name)|Rest], 
    PreviousTryNexts, F, A, Nth, NRule, Nprule, Transformed1, Module) ?-
  Head =.. [F|ArgList],
  make_current_rule_name(F,A,Nth,NRule,Nprule,  CHead, _Arity),
  % CHead is clause head here (i.e. with meta args); Head is constrtraint
  construct_previous_try_next(PreviousTryNexts, A, NRule, 0, CHead),
  fill_current_head(CHead, A, Nth,NRule, ArgList, IndexHeadArgs,MetaArgs), 
  % fill in the arguments for the head
  (Nprule \== 0 -> 
     MetaArgs = [CIndex,In,_Applied] 
   ; MetaArgs = [CIndex,In]
  ),
  append(DeleteHeads, KeepHeads, Partners),
  %DeleteHeads first so that these can be deleted in body
  Clause = (CHead ?- Code0), % Code0 to be filled in
  rule_type(Rule, RuleType),
  construct_main_clause(RuleType, Status, Partners, DeleteHeads, KeepHeads, Head, CHead, Body, Nth, Nprule, PInfo, ComInfo, IndexHeadArgs, MetaArgs, Code0, Vars, Cans, InsChecks, TryNexts0, PreFiredCodes0, PostFiredCodes0, AllRest, Module),
  construct_2ndclause(CHead, Nth, NRule, A, SecondClause, TryNexts1),

  (Partners \== [] -> % not needed if no partners
     construct_prefired_otherpartners(RuleType, Status, Partners, DeleteHeads, KeepHeads, CHead, Head, Body, A, AllRest,
        Nth, NRule, Nprule, Cans, InsChecks, PInfo, ComInfo, PreFiredName, PreFiredCodes1, PostFiredCodes1, TryNexts2, PrePRule, Module),
     construct_postfired_otherpartners(RuleType, Status, Partners, DeleteHeads, KeepHeads, CHead, Head, Body, A, AllRest,
        Nth, NRule, Nprule, Cans, InsChecks, PInfo, ComInfo, PostFiredName, PostFiredCodes2, TryNexts3, PostPRule, Module),
     % fill in various bits of code
     functor(CHead, CFunc, _),
     construct_instancematches(Partners, Cans, Vars, InsChecks, CFunc, 1, Module),

     fill_in_try_otherrest(PreFiredCodes0, PreFiredName),
     fill_in_try_otherrest(PreFiredCodes1, PreFiredName),
     fill_in_try_otherrest(PostFiredCodes0, PostFiredName),
     fill_in_try_otherrest(PostFiredCodes1, PostFiredName),
     fill_in_try_otherrest(PostFiredCodes2, PostFiredName),
     append(TryNexts0,TryNexts1, TryNexts4),
     append(TryNexts4, TryNexts2, TryNexts5),
     append(TryNexts5, TryNexts3, TryNexts),
     PrePRule = [Pre1,Pre2], PostPRule = [Post1,Post2],
     Transformed1 = [Clause,SecondClause,Pre1,Pre2,Post1,Post2|Transformed2]

   ; /* should not have prefirecodes etc. if no partners */
     append(TryNexts0, TryNexts1, TryNexts),
     Transformed1 = [Clause,SecondClause|Transformed2]
  ),


  Nth1 is Nth + 1,
  gen_code_for_constraint(Rest, TryNexts, F, A, Nth1, NRule, Nprule, Transformed2, Module).


construct_instancematches([], [], [], [], _, _, _) :- !.
construct_instancematches([Partner|Ps], [Candidate|Cans], [Common-Remain|Cs], [ICheck|ICs], CFunc, N, Module) :-
   copy_term(Remain,CRemain),
   concat_atom([CFunc,match,N], NewFunc),
   ICheck =.. [NewFunc,Candidate,Common, Remain],
   GoalHead =.. [NewFunc,Partner,Common, CRemain],
   compile_term((GoalHead ?- CRemain = Remain))@Module,
   N1 is N + 1, 
   construct_instancematches(Ps, Cans, Cs, ICs, CFunc, N1, Module).


fill_in_try_otherrest([], _Name) :- !.
fill_in_try_otherrest([f(ArgList,TobeFilled)|Others], Name) :-
   (var(TobeFilled) -> TobeFilled =.. [Name|ArgList] ; true),
   % TobeFilled is not var if no code needs to be filled in
   fill_in_try_otherrest(Others, Name).


rule_type((_H==>_B), Type) ?-  !, Type = propagation.
rule_type((H<=>_B), Type) ?-
  H = (_\_) ->
     Type = simpogation ; Type = simplification.


construct_main_clause(RuleType, Status, Partners, DeleteHeads, KeepHeads, Head, CHead, Body, Nth, Nprule, PropInfo, CompInfo, 
     IndexHeadArgs, MetaArgs, Code0, Vars, Candidates, 
     InstanceChecks, TryNexts, PreFiredCodes, PostFiredCode, AllRest, Module) :-

  get_constraintnumber(Head, ConstNo, Module),
  decompose_body(Body, Guard, BGoals),

  % construct the clause
  CommonArgs = a(Head,CHead,MetaArgs,Nth,AllRest,_PIndecies,_GConst,
     _Add_Applied,AppRest),
  ExtendArgs = a(Status,ConstNo,BGoals,DeleteHeads,KeepHeads,PostFired0,NTryNext),

  construct_find_partners(Partners, CommonArgs, 1, Code0, TryNexts0, Vars, Candidates, InstanceChecks, Code1, Module),
  construct_applied(RuleType, Partners, CommonArgs, PropInfo, CompInfo, Code1, 
     PreFired0, TryNext0, Code2),
  construct_guard(Guard, (Head,Partners), CommonArgs, CompInfo, ExtendArgs, Nprule, Code2, PreFired1, TryNext, Code3, Module),
  construct_body(CommonArgs, ExtendArgs, Nprule, Code3, PostFired, TryNext1, Module),

  % construct the various arg list
  multi_append([TryNext0,TryNext,TryNext1,TryNexts0,NTryNext], TryNexts),
  append(IndexHeadArgs, [AllRest], L0),
  append(L0, MetaArgs, Post1),
  append(L0, [applied|MetaArgs], Pre0), append(L0, [gf|MetaArgs], Pre1),
  (Nprule == 0 ->
      Post1 = PostFiredArgs,
      Pre1 = PreArgs1,
      Pre0 = PreArgs0 
    ; append(Post1, [AppRest], PostFiredArgs),
      append(Pre0, [AppRest], PreArgs0),
      append(Pre1, [AppRest], PreArgs1)
  ),
  PostFiredCode = [f(PostFiredArgs,PostFired),f(PostFiredArgs,PostFired0)],
  PreFiredCodes = [f(PreArgs0,PreFired0),f(PreArgs1,PreFired1)].


construct_2ndclause(OldHead, Nth, NRule, ConsArity, Code, TryNexts) :-
   functor(OldHead, F,A), 
   functor(NewHead, F,A), % construct new head
   InPos is ConsArity + NRule + 2, % second meta-arg is In
   arg(InPos, NewHead, In),
   (NRule > 0 ->
      arg(Nth, NewHead, I), % the indexing var. for this clause
      %% MAGIC NUMBER
      TryNexts = [f(TryNext1,[Nth-hf|In],NewHead),f(TryNext0,In,NewHead)],
      Code = (NewHead :-
         I == 1 -> TryNext0 
	      ; TryNext1
      )
    ; Code = (NewHead :- TryNext0),
      TryNexts = [f(TryNext0,[Nth-hf|In],NewHead)]
   ).


fill_previous_infoargs(PInfoPos, NRule, TryOtherHead, [FailType,SuspL,AppRest]) :-
   arg(PInfoPos, TryOtherHead, FailType),
   SuspPos is NRule + 2,
   arg(SuspPos, TryOtherHead, SuspL),
   AppRestPos is NRule + 3,
   arg(AppRestPos, TryOtherHead, AppRest).


construct_previous_try_next([], _, _, _, _) :- !.
construct_previous_try_next([f(TryNext, NewIn, PrevHead)|TryNexts], OArity, NRule, RemoveArg, NewHead) :-
% RemoveArg normally set to zero, so that indexing args are not removed, except for last clause for this constraint
   (TryNext \== true ->
     functor(NewHead, Name, _A), % NewHead used as template to construct calling goal
     add_inarg(PrevHead, OArity, NRule, NewIn, NewArgs0),
     remove_indexing(RemoveArg, NewArgs0, NewArgs),
     TryNext =.. [Name|NewArgs] % code for calling goal
    ;true
   ),
   construct_previous_try_next(TryNexts, OArity, NRule, RemoveArg, NewHead).

/* updates MetaArgs from OldIn to NewIn */
add_inarg(PrevHead, OArity, NRule, NewIn, NewArgs) :-
   divide_vars(PrevHead, OArity, NRule, OIndexHeadArgs, OMetaArgs),
   update_inarg(OMetaArgs,NewIn, NMetaArgs),
   append(OIndexHeadArgs, NMetaArgs, NewArgs).

/* updates In when given Metaargs */
update_inarg([Index,_OldIn|Rest], NewIn, [Index,NewIn|Rest]).


/* removes the first Nth arguments from rule clause, which are for indexing */
remove_indexing(0, Args0, Args1) :- !, Args0 = Args1. % nothing to remove
remove_indexing(1, [_|Args0], Args1) :- !, Args0 = Args1.
remove_indexing(N, [_|Args0], Args1) :- N > 1,
   N1 is N - 1,
   remove_indexing(N1, Args0, Args1).
 

/* connect_orig_args(+OrigHead, +NRule, +IndexHeadArgs)
   connect the original args. (i.e. "real" arguments from the constraint 
   OrigHead) with their corresponding args. in the current clause being
   constructed -- these are in the IndexHeadArgs, which includes the 
   indexing args at the start
*/
connect_orig_args(OHead, NRule, IndexHeadArgs) :-
   OHead =.. [_|RealArgs],
   remove_indexing(NRule, IndexHeadArgs, RealArgs).



construct_prefired_otherpartners(RuleType, Status, Partners, DeleteHeads,KeepHeads, 
  RuleF, Head, Body, OrigArity, AllRests, Nth, NRule, Nprule, Candidates, CheckInsts, PInfo, ComInfo, PreFiredName, 
  PreFiredCodes, PostFiredCode, TryNexts, Code0, Module) :- 
   decompose_body(Body, Guard, BGoals),
   get_constraintnumber(Head, ConstNo, Module),
   make_try_other_name(RuleF, prefired, PreFiredName),
   nmetaargs(Nprule, BasicSize), % number of "basic" meta-args
   (Nprule == 0 -> MetaSize is BasicSize  + 2; MetaSize is BasicSize + 3),
   % 4 fixed meta-args (AllRest,Reason,Index,In) + 2 w/propagation (Applied, AppRest)
   Arity is NRule + OrigArity + MetaSize,
   functor(CHead, PreFiredName, Arity),
   divide_vars(CHead, OrigArity, NRule, IndexHeadArgs, PreMetaArgs),
   connect_orig_args(Head, NRule, IndexHeadArgs),
   PreMetaArgs = [AllRests,_Reason|MetaArgs],
   get_appliedrest(Nprule, MetaArgs, AppRest),
   Code0 = [(CHead ?- Code1), (CHead :- NextRule)],
   MetaArgs = [_Index,In|_],
true,
   construct_basic_head(CHead, Head, NRule, Nprule, MetaArgs, ConsHead),
   TryNext0 = [f(NextRule, In, ConsHead)],
   CommonArgs = a(Head,CHead,MetaArgs,Nth,AllRests1,_PIndecies,_GConst,
     _Add_Applied, NAppRest),
   ExtendArgs = a(Status,ConstNo,BGoals,DeleteHeads,KeepHeads,PostFired0,NTryNext),

   construct_other_findpartners(AllRests, Partners, CommonArgs, NRule, 
      1, Nprule, Candidates, CheckInsts, Code1, Code2),
   construct_fapplied(RuleType, AppRest, CommonArgs, PInfo, ComInfo, Code2, 
      PreFired0, Code3),
   construct_fguard(Guard, (Head,Partners), CommonArgs, ExtendArgs, ComInfo, Nprule, Code3, Code4, Module),
   construct_body(CommonArgs, ExtendArgs, Nprule, Code4, PostFired, TryNext1, Module),

   % same args as head, except for AllRest, Reason and AppRest
   % no need to append TryNext0, should not have pre/post fired if no 
   % partner
   new_metaargs(Nprule, MetaArgs, NAppRest, NewMetaArgs),
   multi_append([TryNext0,TryNext1,NTryNext], TryNexts),
   append(IndexHeadArgs, [AllRests1,applied|NewMetaArgs], PreArgs0),
   append(IndexHeadArgs, [AllRests1|NewMetaArgs], PostArgs),
   PreFiredCodes = [f(PreArgs0,PreFired0)],
   PostFiredCode =  [f(PostArgs,PostFired),f(PostArgs,PostFired0)].



construct_postfired_otherpartners(RuleType, Status, Partners, DeleteHeads, KeepHeads, RuleF, Head, Body, OrigArity, AllRests, Nth, NRule, Nprule, Candidates, CheckInsts, PInfo, ComInfo, PostFiredName, PostFiredCodes, TryNexts, Code0, Module) :-
   decompose_body(Body, Guard, BGoals),
   get_constraintnumber(Head, ConstNo, Module),
   make_try_other_name(RuleF, postfired, PostFiredName),
   nmetaargs(Nprule, BasicSize),
   (Nprule ==0 -> MetaSize is BasicSize +1 ; MetaSize is BasicSize + 2),
   % should try to remove these magic numbers
   Arity is NRule + OrigArity + MetaSize,
   functor(CHead, PostFiredName, Arity),
   divide_vars(CHead, OrigArity, NRule, IndexHeadArgs, PostMetaArgs),
   connect_orig_args(Head, NRule, IndexHeadArgs),
   PostMetaArgs = [AllRests|MetaArgs],
   get_appliedrest(Nprule, MetaArgs, AppRest),
   Code0 = [( CHead :- Code1), (CHead :- NextRule)],
   MetaArgs = [_Index,In|_],
true,
   construct_basic_head(CHead, Head, NRule, Nprule, MetaArgs, ConsHead),
   TryNext0 = [f(NextRule, In, ConsHead)],
   CommonArgs = a(Head,CHead,MetaArgs,Nth,AllRests1,_PIndecies,_GConst,
     _Add_Applied,AppRest1),
   ExtendArgs = a(Status,ConstNo,BGoals,DeleteHeads,KeepHeads,PostFired3,NTryNext),


   construct_other_findpartners(AllRests, Partners, CommonArgs, NRule,  
     1, Nprule, Candidates, CheckInsts, Code1, Code2),
   construct_fapplied(RuleType, AppRest, CommonArgs, PInfo, ComInfo, Code2, PostFired0, Code3),
   construct_fguard(Guard, (Head,Partners), CommonArgs, ExtendArgs, ComInfo, Nprule, Code3, Code4, Module),
   construct_body(CommonArgs, ExtendArgs, Nprule, Code4, PostFired2, TryNext1, Module),
   % no need to delete post-fired

   multi_append([TryNext0,TryNext1,NTryNext], TryNexts),
   
   (Status == keep -> true; fill_in_with_true(TryNexts)),
   new_metaargs(Nprule, MetaArgs, AppRest1, NewMetaArgs),
   append(IndexHeadArgs, [AllRests1|NewMetaArgs], PostArgs),
   PostFiredCodes = [f(PostArgs,PostFired0),f(PostArgs,PostFired2),f(PostArgs,PostFired3)].


/* new_metaargs(+Nprule, +OldArg, -NewAppRest, NewArgs)
   creates NewArgs, the new meta-args, for the recursive calls to pre- and post-
   fired clauses, from within a pre- and post- fire clause. OldArgs are the
   original meta-args upon entry into the pre-/post-fired clause that is 
   being created, and the only value that needs to be changed is the AppRest
   argument, if it exist
*/
new_metaargs(Nprule, OldArgs, NewAppRest, NewArgs):-
    (Nprule =:= 0 -> NewArgs = OldArgs ; 
         OldArgs = [Index,In,Applied,_],
         NewArgs = [Index,In,Applied,NewAppRest]
    ).


fill_in_with_true([]) :- !.
fill_in_with_true([f(ToBeTrue, _, _)|TryNexts]) :- 
   ToBeTrue = true,
   fill_in_with_true(TryNexts).
      


/* fill_current_head(+CurrentHead, +OriginalArity, +Nth, +NRule, +ArgList,
      -IndexHeadArgs, -MetaArgs)
   CurrentHead is the head of the translated Prolog goal for the current
   rule being compiled. OriginalArity is the
   original arity of the head constraints that triggered the trying of the
   current rule, Nth indicates that the current rule is the Nth rule for this 
   constraint, and is used to construct the indexing, NRule is the number of 
   rules for this constraint, which determines how many arguments will be 
   needed for indexing. ArgList is a list of the original arguments in the 
   constraint. IndexHeadArgs will be returned with the indexing and head 
   (original arguments of the constraint) arguments of CurrentHead, and
   MetaArgs are the "meta" arguments of CurrentHead, i.e. those needed to
   allow for the suspension and rewakening of the constraint.
*/
fill_current_head(CHead, A, Nth, NRule, ArgList, IndexHeadArgs, MetaArgs) ?-
/* current scheme for head args are 
   (Indexing..., Original Args..., ConstraintIndex, FailureInfo, Applied) */
   (NRule \== 0 -> arg(Nth, CHead, 1) ; true),% indexing
   fill_original_args(0, A, NRule, ArgList, CHead),
   divide_vars(CHead, A, NRule, IndexHeadArgs, MetaArgs).


/* divide_vars(+CurrentHead, +OrigArity, +NRule, -IndexHeadArgs, -MetaArgs)
   splits the arguments of CurrentHead into two lists: IndexHeadArgs, which
   are the Indexing and head args, and MetaArgs, which are the meta args
*/
divide_vars(CHead, OArity, NRule, IndexHeadArgs, MetaArgs) :-
   CHead =.. [_|AllArgs],
   IndexHeadSize is OArity + NRule,
   split_list(IndexHeadSize, AllArgs, IndexHeadArgs, MetaArgs).

fill_original_args(N, N, _, [], _) :- !.
fill_original_args(N0, Max, NRule, [Arg|ArgList], CHead) :-
   N1 is N0 + 1,
   Pos is  NRule + N1,
   arg(Pos, CHead, Arg), % put Arg in
   fill_original_args(N1, Max, NRule, ArgList, CHead).


/* get_appliedrest(+Nprule, +MetaArgs, -AppRest)
   returns the AppRest meta-argument from MetaArgs if it exists (if there are no
   propagation rules, i.e. Nprule = 0). This is for the pre and post fired
   clauses, where the two propagation args would be at the end if they exist
   The clause simply does not no anything if AppRest does not exist -- this 
   means that following clauses should not make use of AppRest
*/
get_appliedrest(Nprule, [_,_|Pos], AppRest) :-
   (Nprule =\= 0 -> Pos = [_Applied,AppRest] ; true). 

/* get_constraintnumber(+Constraint, -ListNumber, +Module)
   returns the ListNumber for constraint, i.e. which constraint list a 
   particular constraint (in the form of a structure) is stored in
*/
get_constraintnumber(Cons, ConstNo, Module) :-
   functor(Cons,F,A),
   get_constraintnumber_fa(F,A, ConstNo, Module).

get_constraintnumber_fa(F,A, ConstNo, Module) :-
   recorded_list('CHRconstraints', Constraints)@Module,
   second_member(Constraints, F/A, ConstNo).


construct_check_identical(Partner, Head, PartnerIndex, CIndex, CheckIdentical) :-
   /*get_constraintnumber(Partner, PartnerIndex),*/
   (\+ \+(Partner = Head) -> 
   % if unifiable, then need to generate code to check Partner is not current
       CheckIdentical = (PartnerIndex \== CIndex) ; CheckIdentical = true
   ).


construct_find_partners([],a(_,_,_,_,AllRest,PIndex,_,_,_),_,Code0,NextRules,Com,Cans,InsChk,Code,_M) :- !, 
   Code0 = Code, NextRules = [], AllRest = [], PIndex = [], Com = [], 
   InsChk = [], Cans = [].
construct_find_partners([Partner|Partners],a(Head,CHead,MetaArgs,Nth,AllRest,
   [PartnerIndex|PIndecies],_,_,_), Cth,
   Code, [f(Try_Next_Rule,InsertFCond,CHead)|OtherNextRules], [Common-Remain|Cs],
   [Candidate|Cans], [InstanceCheck|ICs], RestCode, Module) :-
% Nth rule for this constraint, Cth partner, In for failure info
% Head is the raw constraint head (i.e. without meta args)
   MetaArgs = [CIndex,In|_],
   shared_vars(Partner, Head, Common, Remain),
   % non-shared variables in Candidate and Partner are the same 
   get_constraintnumber(Partner,ConstNo, Module),
   (Common \== [] ->
      GetSList = call_explicit(get_constraint_list(Common, ConstNo, SuspL), ech)

    ; GetSList = call_explicit(get_global_constraint(ConstNo, SuspL), ech)
   ),
   construct_check_identical(Partner, Head, PartnerIndex, CIndex, CheckIdentical),
   Code = (
      GetSList,
      (call_explicit(
          find_partner(SuspL, PartnerIndex, Candidate, Rest, Condition), ech),
      InstanceCheck,
       CheckIdentical,
       Condition == yes ->
            RestCode0
          ; !, Try_Next_Rule
      )
   ),
   AllRest = [Rest|AllRest0],
   InsertFCond = [Nth-Cth-np|In],
   Cth1 is Cth + 1,
   construct_find_partners(Partners,a(Head,CHead,MetaArgs,Nth,AllRest0,
      PIndecies,_,_,_), Cth1, RestCode0, OtherNextRules, Cs, Cans, ICs,
      RestCode,Module).

/* construct_basic_head(+CurrentClauseHead, +OrigConstraintHead, +NRule,
       +Nprule, +BasicMetaArgs, -ConstructedHead)
   constructs a "basic" head with just the indexing, real arguments, and
   the basic meta-args. CurrentClauseHead is the head of the clause that
   is being constructed. This can contain additional meta-args to the basic
   one. These need to be removed. OrigConstraintHead is the constraint,
   without any indexing or meta-args. NRule is the number of rules for this
   constraint (i.e. number of indexing args.). BasicMetaArgs is the basic
   meta-args for the CurrentClauseHead
*/
construct_basic_head(CHead, Head, NRule, Nprule, MetaArgs, NewHead) :- 
   functor(Head,F,A),
   divide_vars(CHead, A, NRule, IndexHeadArgs, _),
   nmetaargs(Nprule, N), length(MetaArgs1, N),
   append(MetaArgs1, _, MetaArgs), 
   % get rid of any extra non-basic arguments at end of MetaArgs
   append(IndexHeadArgs, MetaArgs1, Args),
   NewHead =.. [F|Args].


construct_other_findpartners([], [], a(_,_,_,_,AllRest,PIndex,_,_,_), _, _, _,
      _, _, Code0, Code) :- !,
   Code0 = Code, AllRest = [], PIndex = [].
construct_other_findpartners([Rest|Rests], [Partner|Partners], a(Head,CHead, 
    MetaArgs,Nth,AllNRest,[PartnerIndex|PIndecies],_,_,_), NRule, Cth, Nprule,
    [Candidate|Cans], [CheckIns|CIs], Code, RestCode) :-
   MetaArgs = [CIndex, _In|_],

   % construct ConsHead which contains only the basic meta-args
   %construct_basic_head(CHead, Head, NRule, Nprule, MetaArgs, ConsHead),

   construct_check_identical(Partner, Head, PartnerIndex, CIndex, CheckIdentical),
   Code = ( 
      call_explicit(
      ffind_partner(Rest, PartnerIndex, Candidate, NewRest), ech),
      CheckIns,
      CheckIdentical,

      RestCode0
   ),
   AllNRest = [NewRest|NewRests],
   Cth1 is Cth + 1,
   construct_other_findpartners(Rests, Partners, a(Head,CHead,MetaArgs,Nth,NewRests,PIndecies,_,_,_), NRule,  
     Cth1, Nprule, Cans, CIs, RestCode0, RestCode). 


/* extract_applied(+MetaArgs, -Applied) extracts the argument for storing
   Applied information from the meta arguments of a clause
*/
extract_applied([_,_,Applied|_], Applied).

/* construct_applied_check 
 
 this constructs the appropriate direct applied (i.e. applied list has
 already been extracted) check for propagation rules. This occurs in the
 pre-fired and post-fired clauses. Currently it only deals with two headed
 propagation rules (no check needed for single headed propagation
 rules). In such cases, there are two factors: if the head constraints have
 the same functor/arity, and if the one initiating the rule has already
 been added to the constraint store or not. These generate four different
 situations 
*/

construct_applied_check(1, double(H1,_H2), Head, RuleNo, PList, [PIndex], Index, Add_Applied, Not_AppliedCheck, AppRest, GConsNo, Status) :- !,
	get_currentheadpos(H1, Head, Pos, OtherPos),
	Not_AppliedCheck = (
	    suspension_to_goal(PIndex, PGoal,_),
            arg(constraintnum_pos, PGoal, PConsNumber), % get the ConsNo for partner
            PList = [_|PListRest],
	    (var(Index) -> /* if var, then initiating head is not in constraint store */
	        /* can only check in that case */
	        call_explicit(check_samepairapplied_direct(PList, PListRest, Pos-PConsNumber, Tail, AppRest, Status), ech)

	      ; /* can check and mark if initiating head already in constraint store */
	        suspension_to_goal(Index, Cons0, _),
                arg(constraintnum_pos, Cons0, GConsNo),
              
	        call_explicit(check_samepairapplied_directmarked(PList, PListRest, Pos-PConsNumber, AppRest, OtherPos-GConsNo, PGoal, RuleNo, Status), ech)
            )

        ),
	Add_Applied = samepairapplied(Pos,PConsNumber,Tail,AppRest,OtherPos,PGoal,RuleNo).
construct_applied_check(1, _, _, RuleNo, PList, [PIndex], Index, Add_Applied, 
    Code, AppRest, GConsNo, Status) :- !,
	Code = (suspension_to_goal(PIndex, PGoal,_),
          arg(constraintnum_pos, PGoal, PConsNumber),
          PList = [_|PListRest],
	  (var(Index) ->
                call_explicit(check_pairapplied_direct(PList, PListRest, PConsNumber, Tail, AppRest, Status), ech)

	       ;suspension_to_goal(Index, Cons0, _),
                arg(constraintnum_pos, Cons0, GConsNo),

	        call_explicit(check_pairapplied_directmarked(PList, PListRest, PConsNumber, AppRest, GConsNo, PGoal, RuleNo, Status), ech)
          )
        ),
        Add_Applied = pairapplied(PConsNumber,Tail,AppRest,PGoal,RuleNo).
construct_applied_check(N, _, _, _RuleNo, _PList, _, _Index, Add_Applied, 
    Code, _AppRest, _, _Status) :- !,
   N > 1,
   writeln(error, "Propagation rule with more than two head constraints not yet supported. No applied testing done.\n"), flush(error),
   Add_Applied = true,
   Code = true.





appropriate_applied(0, _, _, RuleNo, [], Applied, Code, Add_Applied, 
  AppRest, _ConsNo, Status) :- !,
    Code = ( nonvar(Applied) ->
                arg(RuleNo, Applied, App),
	        (App == * -> Status = found ; Status = notfound)
	     ; % var(Applied)
               number_of_propagations(Size), 
               call_explicit(create_applied(Applied, Size),ech),
               Status = notfound
           ), AppRest = [],
    Add_Applied = arg(RuleNo, Applied, *).   % * mark as read
appropriate_applied(1, double(H1, _H2), Head, RuleNo, [PIndex], Applied, Code, 
  Add_Applied, AppRest, ConsNo, Status) :- !,
    % propagation rule, with two heads that has the same functor and arity
    % needs special check and insert applies
    get_currentheadpos(H1, Head, Pos, OtherPos),
    Code = ( suspension_to_goal(PIndex, PGoal, _),
              arg(constraintnum_pos, PGoal, PConsNumber),
              call_explicit(check_samepairapplied(RuleNo, Applied, Pos-PConsNumber, Tail, AppRest, Status), ech)
     ),
     Add_Applied = call_explicit(insert_samepairapplied(Pos-PConsNumber, Tail, AppRest, OtherPos-ConsNo, PGoal, RuleNo), ech).
appropriate_applied(1, _, _, RuleNo, [PIndex], Applied, Code, Add_Applied, AppRest, ConsNo,
   Status) :- !,
     Code = ( suspension_to_goal(PIndex, PGoal, _),
              arg(constraintnum_pos, PGoal, PConsNumber),
              call_explicit(check_pairapplied(RuleNo, Applied, PConsNumber, Tail, AppRest, Status), ech)
     ),
     Add_Applied = call_explicit(insert_pairapplied(PConsNumber, Tail, AppRest, ConsNo, PGoal, RuleNo), ech).
appropriate_applied(N, _, _, _RuleNo, _, _, Code, Add_Applied, _, _, _) :- 
   N > 1,
   writeln(error, "Propagation rule with more than two head constraints not yet supported. Not applied testing done.\n"), flush(error),
   Code = true,
   Add_Applied = true.



/* get_currentheadpos(+Head1, +Partner, -Pos, -OtherPos) 
   returns in Pos the position (1 or 2) of the current head that is being
   transformed. OtherPos is the other position. Head1 is the first (leftmost)
   head as occur in the rule, and Head is the current active head
*/
get_currentheadpos(Head1, Head, Pos, OPos) :-
   Head1 \== Head -> Pos = 2, OPos = 1; Pos = 1, OPos = 2.




construct_applied(Rule_Type, Partners, a(Head,CHead,MetaArgs,Nth,_,PIndecies,
   GConsNo,Add_Applied,AppRest), RuleNo, ComInfo, Code, PreFiredAlt, TryNext, 
   RestCode) :-
   Rule_Type == propagation ->
     length(PIndecies, NPartners),
     extract_applied(MetaArgs, Applied),
     appropriate_applied(NPartners, ComInfo, Head, RuleNo, PIndecies, Applied, Not_Applied, Add_Applied, AppRest, GConsNo, Status),

     Code = (
        Not_Applied,
        (Status == notfound ->
            RestCode
	  ; AltAction
        )
     ),
     (Partners == [] ->
         MetaArgs = [_Index,In|_], % get In
         TryNext = [f(AltAction, [Nth-notapp|In], CHead)],
         PreFiredAlt = true
       ; TryNext = [], PreFiredAlt = AltAction
     )
   ; Add_Applied = true, PreFiredAlt = true, TryNext = [],
     RestCode = Code.
 
construct_fapplied(Rule_Type, Partners, a(Head,_,MetaArgs,_,_,PIndecies,GConsNo,Add_Applied,AppRest), RuleNo, ComInfo, Code, PreFiredAlt, RestCode) :-
   Rule_Type == propagation ->
     MetaArgs = [Index|_],
     length(PIndecies, NPartners),
     construct_applied_check(NPartners, ComInfo, Head, RuleNo, Partners, PIndecies, Index, Add_Applied, Not_AppliedCheck, AppRest, GConsNo, Status),

     Code = (
        Not_AppliedCheck,
        (Status == notfound ->
            RestCode
           ; PreFiredAlt
        )
     )
   ; 
     Add_Applied = true,
     RestCode = Code.

/* metaguard_code(+Guard, -Code)
   construct the code for "meta" Guards -- i.e. none of the guards can 
   constrain the variables in them
*/
metaguard_code(Guard, Code) :-
    Code = (last_suspension(Mark),
            Guard, ttrue, % need to mark differently so not optimised away
            new_suspensions(Mark,[]) % no goals in guard has been delayed
           ).

/* guard_code(+Guard, +Global, -Code)
   construct the code for Guards, where they contain goals which are capable
   of constraining (global)variables. Code taken from old chr.pl:

% Before the guard (Goal) is called, a 'fail' is attached to every variable
% of the Goal. Then, as soon as one of these variables is touched
% (unified), the call will fail.

*/
guard_code(Guard, Globals, Code) :-
    Code = (
      make_suspension(fail, 1, Susp),
      insert_suspension(Globals, Susp, constrained of suspend, suspend),
      last_suspension(Mark),
      Guard, ttrue,
      new_suspensions(Mark, []),
      kill_suspension(Susp)
    ).

construct_invertable_rule(Globals, CommonArgs, InvBody, ExtendArgs, Nprule, Code, Try_Other_Partners, TryNext, Module) :- 
    decompose_body(InvBody, InvG, InvB),
    construct_guard(InvG, Globals, CommonArgs, nil, ExtendArgs, Nprule, Code, Try_Other_Partners, TryNext, RestCode, Module),
    ExtendArgs = a(delete,ConstNo,_,DeleteHead,KeepHead,PostFired,TryNext1),
    % Delete and Keep Heads swapped
    construct_body(CommonArgs, a(keep,ConstNo,InvB,KeepHead,DeleteHead,
        PostFired,TryNext1), Nprule, RestCode, PostFired, TryNext1, Module).

construct_guard(Guard, Globals, CommonArgs, ComInfo, ExtendArgs, Nprule, Code, Try_Other_Partners, TryNext, Rest_Code, Module) :-

   CommonArgs = a(_,CHead,MetaArgs,Nth,AllRest,_,_,_,_),  
   (Guard \== true ->
       ((recorded('CHRdont_guard_bindings', Module) ; nonconstrainable_guards(Guard)) ->
       % check if guard can possibly cause the constraining of any variable
           metaguard_code(Guard, Execute_Guard)
         ; guard_code(Guard, Globals, Execute_Guard)
       ),

       (AllRest == [] -> /* no need to try other partners */
           Code = ( 
              Execute_Guard -> !,
                  Rest_Code
	        ; !, AltCode

           ),
           MetaArgs = [_Index,In|_],
           TryNext = [f(Try_Next_Rule, [Nth-gf|In], CHead)],
           Try_Other_Partners = true

         ; %AllRest \== []
           Code = 
             (Execute_Guard -> !,
                 Rest_Code
               ; !, AltCode
           ), 
           TryNext = []
       ),

       (ComInfo = twobodies(InvBody) ->
           construct_invertable_rule(Globals, CommonArgs, InvBody, ExtendArgs, Nprule, AltCode, Try_Other_Partners, TryNext, Module)
 
         ; (AllRest == [] -> AltCode = Try_Next_Rule ; AltCode = Try_Other_Partners),
           ExtendArgs = a(_,_,_,_,_,true,[])
       )
       
     ; /* Guard == true */
       Code = (!, Rest_Code), Try_Other_Partners = true,
       TryNext = []
   ).


construct_fguard(Guard, Globals, CommonArgs, ExtendArgs, ComInfo, Nprule, Code, Rest_Code, Module) :-
   Guard \== true ->
       ((recorded('CHRdont_guard_bindings', Module) ; nonconstrainable_guards(Guard)) ->
       % check if guard can possibly cause the constraining of any variable
           metaguard_code(Guard, Execute_Guard)
         ; guard_code(Guard, Globals, Execute_Guard)
       ),
       (ComInfo = twobodies(Inv) ->
          Code = (Execute_Guard ->
                   !, Rest_Code ; AltCode
                 ),
          construct_finvertable_rule(Globals, Inv, CommonArgs, ExtendArgs, Nprule, AltCode, Module)

        ; Code = ( 
             Execute_Guard, !,
             Rest_Code
          )


       )
     ; /* Guard == true, no difference in twobodies case, as head will be
          removed */
       Code = (!, Rest_Code).



construct_finvertable_rule(Globals, Inv, CommonArgs, ExtendArgs, Nprule, Code, Module) :-
     decompose_body(Inv, InvG, InvB),
     construct_fguard(InvG, Globals, CommonArgs, ExtendArgs, nil, Nprule, Code, RestCode, Module),
     ExtendArgs = a(delete,ConstNo,_,DeleteHead,KeepHead,PostFired,TryNext1),
     % Delete and Keep Heads swapped
     construct_body(CommonArgs, a(keep,ConstNo,InvB,KeepHead,DeleteHead,
        PostFired,TryNext1), Nprule, RestCode, PostFired, TryNext1, Module).



construct_deletion([], _, _, true) :- !.
construct_deletion([DHead|DHs], [Index|Indecies], Constraints, Code) :-
   functor(DHead,F,A),
   second_member(Constraints, F/A, ConstNo), 
   Code = (call_explicit(kill_constraint(ConstNo, Index),ech), Code1), 
   construct_deletion(DHs, Indecies, Constraints, Code1).


construct_body(a(Head,CHead,MetaArgs,_,AllRest,PIndecies,GConst,Add_Applied,_),
  ExtendArgs,  Nprule, Code, PostFired, TryNext, Module) :-

   ExtendArgs = a(HeadAction,ConstNo,BodyGoals,DeleteHeads,_,_,_),
   recorded_list('CHRconstraints', Constraints)@Module,
   construct_deletion(DeleteHeads, PIndecies, Constraints, DeleteConstraints),
   MetaArgs = [Index,In|_], 
   nmetaargs(Nprule, BasicSize),
   length(BasicMetaArgs, BasicSize),  % only basic MetaArgs are used in suspended constraint
   append(BasicMetaArgs, _, MetaArgs),
   (HeadAction == keep ->
       construct_wrapper_constraint(Head, BasicMetaArgs, Constraint, GConst),
       generate_constraint_suspend(Add_Applied, Index, Constraint, Head, ConstNo, GConst, SCode),
       Code = (
          SCode,
          DeleteConstraints,
          BodyGoals, 
          May_Continue
       ),
       (AllRest == [] ->
           TryNext = [f(May_Continue,In,CHead)],
           PostFired = true
	;  TryNext = [],
           (kill_not_possible(BodyGoals) -> 
           % if body goals could not possibly cause constraint to be killed,
           % then no need to test if suspension is valid or not.

               May_Continue = PostFired

             ; May_Continue =
                (is_suspension(Index) ->
	            PostFired ; true
                )
           )
       )

     ; /* HeadAction \= keep */
       Code = (
         call_explicit(kill_constraint(ConstNo, Index),ech),
         DeleteConstraints,
         BodyGoals
       ), PostFired = true, TryNext = []
   ).


/* generate the code for dealing with the possible addition of a constraint
   to the constraint store after a rule is fired. If the rule is a two
   headed propagation rule, then special code may need to be generated to
   deal with checking if the heads have been applied already. There is also
   different code in this case for the initial call and the pre- and post-
   fired calls
*/
generate_constraint_suspend(samepairapplied(Pos,PConsNumber,Tail,AppRest,
    OtherPos,PGoal,RuleNo), _Index, Constraint, Head, ConstNo, GConst, Code) ?- !,
% double headed propagation rule, with same (same functor) heads. Here 
% Add_Applied (first arg.) is used to pass extra args for constructing code
	Code = 
          (var(GConst) -> % check GConst instead of Index. Previous check of var(Index) means GConst would be instatiated if Index non-var
	      call_explicit(suspend_constraint(Constraint, Head, ConstNo), ech),
	      call_explicit(insert_samepairapplied(Pos-PConsNumber, Tail, AppRest, OtherPos-GConst, PGoal, RuleNo), ech)

            ; true
          ).
generate_constraint_suspend(pairapplied(PConsNumber,Tail,AppRest,PGoal,RuleNo),
    _Index, Constraint, Head, ConstNo, GConst, Code) ?- !,
% double headed propagation rule, with different heads
	Code = 
          (var(GConst) ->
	      call_explicit(suspend_constraint(Constraint, Head, ConstNo), ech),
	      call_explicit(insert_pairapplied(PConsNumber, Tail, AppRest, GConst, PGoal, RuleNo), ech)

            ; true
          ).
generate_constraint_suspend(Add_Applied, Index, Constraint, Head, ConstNo, GConst, Code) :- 
% generic case
       Code = ( 
          (var(Index) ->
	      call_explicit(suspend_constraint(Constraint, Head, ConstNo), ech)
            ; suspension_to_goal(Index, Cons0, _),
              arg(constraintnum_pos, Cons0, GConst)
              %Cons0 is Constraint, but use new var as no need to deconstruct
          ),
          Add_Applied
      ).


/* kill_not_possible(+Goal) succeeds if body goal(s) Goal cannot possibly
   cause a constraint to be killed. Currently, just built-ins
*/
kill_not_possible((Goal, Goals)) :- !,
   kill_not_possible(Goal), kill_not_possible(Goals).
kill_not_possible(Call) :-
   call_type_goal(Call), !, fail.
kill_not_possible(Goal) :-
   functor(Goal, F,A),
   get_flag(F/A, type, built_in).

call_type_goal(call(_)) :- !.
call_type_goal(call(_,_)) :- !.
call_type_goal(call_explicit(_,_)) :- !.
call_type_goal(call_priority(_,_)) :- !.
call_type_goal(subcall(_,_)).


/* make_current_rule_name(+CFunctor, +CArity, +Nth, +NRule, +Nprule, -NewRuleHead, -NewArity)
   creates a new clause head (NewRuleHead) for a translated rule called with a 
   constraint CFunctor/CArity. This clause is the Nth clause for this constraint.
   There are a total of NRule for this constraint. 
   Nprule is the number of propagation rule in the program, if this is zero, the
   clause head is simplier as it omits the Applied argument
*/
make_current_rule_name(CF, CA, Nth, NRule, Nprule, RuleHead, Arity) :-
    (Nprule =:= 0 -> Extra = 0 ; Extra = 1),
    Arity is CA + Extra + NRule + 2, % NRule args for indexing; 2 extra meta arguments
    make_rulefunctor(CF, CA, Nth, Functor),
    functor(RuleHead, Functor, Arity).

make_rulefunctor(CF, CA, Nth, Functor) :- % creates a (hopefully) unique new name
    concat_atom(['CHR',CF,CA,'_',Nth], Functor).

/* make_try_other_name(+MainRuleHead, +Position, -Name)
   creates the name for the auxillary predicates for the rule with MainRuleHead
   as head of the rule. Position is either prefired or postfired, depending
   on if the goal is to be called before or after the firing of the rule
*/
make_try_other_name(MainRuleHead, Position, Name) :-
   functor(MainRuleHead, F, _),
   concat_atom([F,Position], Name).


/* construct_wrapper_constraint(+CHead, +ExtraArgs, -Constraint, -GlobalNumber)
  constructs the actual suspension (Constraint) that is used to represent 
  the suspended constraint CHead. ExtraArgs is a list of the meta arguments
  carried by the actual suspension. GlobalNumber is the global constraint
  number used to index CHR constraints for this Constraint
*/
construct_wrapper_constraint(CHead, ExtraArgs, Constraint, GConsNo) :-
   functor(CHead, CF, CA),
   concat_atom(['CHRsusp',CF,CA], WFunctor), 
   Constraint =.. [WFunctor, GConsNo, CHead|ExtraArgs].
   
nmetaargs(Nprule, N) :-
% Currently MetaArgs are Index,In,Applied
    (Nprule =:= 0 -> N = 2 ; N = 3).


optimise([], []) :- !.
optimise([Clause|Code], [OClause|OCode]) :-
   optimise_clause(Clause, OClause),
   optimise(Code, OCode).

optimise_clause((:- Goal), (:- Goal)) :- !.
optimise_clause((Head :- Body), (Head :- OBody)) :- !,
   optimise_body(Body, OBody0), optimise_further(OBody0, OBody).
optimise_clause((Head ?- Body), (Head ?- OBody)) :- !,
   optimise_body(Body, OBody0), optimise_further(OBody0, OBody).
optimise_clause(Fact, Fact).


% convert ttrue to true and remove singleton trues that may be left.
optimise_further(ttrue, Out) ?- !, Out = true.
optimise_further((ttrue, Goals), Out) ?- !, Out = (true,OGoals),
   optimise_further(Goals, OGoals).
optimise_further((true, Goals), OGoals) ?- !,
   optimise_further(Goals, OGoals).
optimise_further((Goal, true), OGoal) ?- !,
   optimise_further(Goal, OGoal).
optimise_further((Goal,Goals), Out) ?- !, Out = (OGoal,OGoals),
   optimise_further(Goal, OGoal), optimise_further(Goals, OGoals).
optimise_further((Goal1;Goal2), Out) ?- !, Out = (OGoal1;OGoal2),
   optimise_further(Goal1, OGoal1), optimise_further(Goal2, OGoal2).
optimise_further((If -> Then), Out) ?- !, Out = (OIf -> OThen),
   optimise_further(If, OIf), optimise_further(Then, OThen).
optimise_further(Goal, Goal).


optimise_body((true, Goals), OGoals) ?- !,
   optimise_body(Goals, OGoals).
optimise_body((Goal,true), OGoal) ?- !,
   optimise_body(Goal, OGoal).
optimise_body((Goal,Goals), Out) ?- !, Out = (OGoal,OGoals),
   optimise_body(Goal, OGoal),
   optimise_body(Goals, OGoals).
optimise_body((Goals1;Goals2), Out) ?- !, Out = (OGoals1;OGoals2),
   optimise_body(Goals1, OGoals1),
   optimise_body(Goals2, OGoals2).
optimise_body((If -> Then), Out) ?- !, Out = (OIf -> OThen),
   optimise_body(If, OIf),
   optimise_body(Then, OThen).
optimise_body(Goal, Goal).


printcode([]) :- !.
printcode([Clause|Clauses]) :-
   writeclause(user,Clause), nl,
   printcode(Clauses).


/* nonconstrainable_guards(+Guards) succeeds if all guard goals in Guards
   cannot possibly constrain the value of their arguments. 
*/
nonconstrainable_guards((Guard1, Guard2)) ?- 
      nonconstrainable_guards(Guard1), nonconstrainable_guards(Guard2).
nonconstrainable_guards(Guard) :- % single guard goal
      nonconstraining_goal(Guard).

nonconstraining_goal(_ > _) :- !.
nonconstraining_goal(_ < _) :- !.
nonconstraining_goal(_ >= _) :- !.
nonconstraining_goal(_ =< _) :- !.
nonconstraining_goal(var(_)) :- !.
nonconstraining_goal(nonvar(_)) :- !.
nonconstraining_goal(_ == _) :- !.
nonconstraining_goal(_ =:= _) :- !.
nonconstraining_goal(_ =\= _) :- !.
nonconstraining_goal(_ \== _) :- !.
nonconstraining_goal(ground(_)) :- !.
nonconstraining_goal(nonground(_)) :- !.
nonconstraining_goal(free(_)) :- !.

/* shared_vars(?T1, ?T2, -Shared, -Remain1) 
     returns in Shared the shared variables of Terms T1 and T2, Remain1 
     will contain the non-shared variables in T1
*/
shared_vars(T1, T2, Shared, Remain1) :-
   term_variables(T1, Vars1),
   term_variables(T2, Vars2),
   intersect(Vars1, Vars2, Shared, Remain1).

intersect([], _, L, R) :- !, L = [], R =[].
intersect([V1|L1], L2, Intersect, Remain) :-
   (membervar(L2, V1) ->
      Intersect = [V1|Intersect0], Remain = Remain0
    ; Intersect = Intersect0, Remain = [V1|Remain0]
   ), intersect(L1, L2, Intersect0, Remain0).


membervar([V1|L], V) :- 
    V == V1 -> true ; membervar(L,V).


initialise_processed(N, Processed) :- 
    functor(Processed, rules, N),
    make_empty_lists(N, Processed).

/* make_list(+CommaList, -List)
   changes CommaList (in the form (A1,A2,....An) to a normal List)
*/
make_list((A1,A2), [A1|L]) :-
    !,
    make_list(A2, L).
make_list(A, [A]).


/* decompose_body(+Body, -GuardGoals, -BodyGoals)
    breaks down the incoming Body into goals for the guard and body
*/
decompose_body((Guards|Goals), Guards, Goals) :- !.
decompose_body(Goals, true, Goals).



syntax_check([], _, _, _).
syntax_check([Rule|Rules], Constraints, PRules, Module) :-
    check_one_rule0(Rule, Constraints, PRules, Module),
    syntax_check(Rules, Constraints, PRules, Module).


check_one_rule0(FRule, Constraints, Processed, Module) :-
    (FRule = (Name ::= Rule) ->
        check_one_rule(Rule, Constraints, Processed, Name, Module)
      ; check_one_rule(FRule, Constraints, Processed, [], Module)
    ).
 
check_one_rule((KeepHeads\DeleteHeads <=> Body), Constraints, Processed, Name,
      Module) ?- !,
    Rule = (KeepHeads\DeleteHeads<=>Body),
    make_list(KeepHeads, KHeadsL0),
    make_list(DeleteHeads, DHeadsL0),
    remove_symmetric(DHeadsL0, DHeadsL),
    %may_remove_symmetric(KHeadsL0, KHeadsL),
    KHeadsL0 = KHeadsL,
    single_simpogation(KHeadsL, DHeadsL, Body, InvBody, Type, Module),
    simpogation_action(Type, Rule, KHeadsL0, KHeadsL, DHeadsL0, DHeadsL, Body, 
       InvBody, Constraints, Processed, Name).
check_one_rule((Heads <=> Body), Constraints, Processed, Name, _Module) ?- !,
    Rule = (Heads <=> Body),
    make_list(Heads, HeadList0), 
    remove_symmetric(HeadList0, HeadList),
    definedheads(HeadList, Constraints, delete, [], HeadList0, Body, 0, Rule, 
      Name, not_prop, nil, Processed).
check_one_rule(PNo-(Heads==>Body), Constraints, Processed, Name, _Module) ?-
    Rule = (Heads==>Body),
    make_list(Heads, HeadList0),
    may_remove_symmetric(Body, HeadList0, HeadList, Info), 
    definedheads(HeadList, Constraints, keep, HeadList0, [], Body, 0, Rule, 
      Name, PNo, Info, Processed).


simpogation_action(not_reducible, Rule, KHeadsL0, KHeadsL, DHeadsL0, DHeadsL, 
  Body, _InvBody, Constraints, Processed, Name) ?- !,

    definedheads(KHeadsL, Constraints, keep, KHeadsL0, DHeadsL0, Body, 0, Rule, 
      Name, not_prop, nil, Processed),
    definedheads(DHeadsL, Constraints, delete, KHeadsL0, DHeadsL0, Body, 0, Rule,
      Name, not_prop, nil, Processed).
simpogation_action(nobody, Rule, KHeadsL0, _KHeadsL, DHeadsL0, DHeadsL, 
  Body, _InvBody, Constraints, Processed, Name) ?- !,
    definedheads(DHeadsL, Constraints, delete, KHeadsL0, DHeadsL0, Body, 0, 
      Rule, Name, not_prop, nil, Processed).
simpogation_action(twobodies, Rule, KHeadsL0, _KHeadsL, DHeadsL0, DHeadsL, 
  Body, InvBody, Constraints, Processed, Name) ?- 
    definedheads(DHeadsL, Constraints, delete, KHeadsL0, DHeadsL0, Body, 0, 
      Rule, Name, not_prop, twobodies(InvBody), Processed).



get_prop_nclausearg(V0,V1) :- var(V0), !, V0 = V1.
get_prop_nclausearg([N-_Head|Rest], [N|Ns]) :-
    get_prop_nclausearg(Rest, Ns).

/* definedheads(+HeadList, +Constraints, +Status, +KeepHeads, +DeleteHeads, 
        +Body, +N, +Rule, +Name, +PropNum, +Info, +Processed)
    definedheads checks if the head constraints in HeadList of rule Rule are 
    defined or not. It also does some initial processing on the rule to ease
    the later translation: the constraints that are kept (KeepHeads) and those
    that are deleted (DeleteHeads) are seperated when called. The HeadList 
    passed is either those constraints that are kept (i.e. same as KeepHeads) 
    or those that are deleted (DeleteHeads), as indicated by Status. Body is
    the body of the rule (guard and body goals). N is used to indicate the Nth
    Head in HeadList is being considered. Processed is used to store the 
    processed rule. For a particular rule, an entry is made in Processed for 
    each head constraint in the rule. Name is the name the user gave to the
    rule, would be [] if not named. PropNum is the propagation rule number for
    this propagation rule (not_prop if not propagation rule). 
    Additional information deduced obtained during the processing which 
    have affected the way processed rules are generated - this information
    is passed along in Info for correct behaviour during transformation

*/
definedheads([H|Hs0], Constraints, Status, AllKHeads, AllDHeads, Body, N0, Rule,  Name, NProp, Info, Processed) ?- 
    functor(H, F, A),
    (second_member(Constraints, F/A, ConsNum) ->
       N1 is N0 + 1,
       (Status == keep ->
	  remove_head(N1, AllKHeads, OtherHeads),
          addto_processed(ConsNum, Rule, keep, H, OtherHeads, AllDHeads, Body, 
             Name, NProp, Info, Processed)
        ; remove_head(N1, AllDHeads, OtherHeads),
          addto_processed(ConsNum, Rule, delete, H, AllKHeads, OtherHeads, 
             Body, Name, NProp, Info, Processed)
       ),
       definedheads(Hs0, Constraints, Status, AllKHeads, AllDHeads, Body, N1, Rule, Name, NProp, Info, Processed)
     ; printf(error, "Syntax error: undefined constraint %a/%w (%w) found in:\n", [F,A,H]),
       flush(error),
       pretty_write(Rule)
       % some error recovery routine
    ). 
definedheads([], _, _, _, _, _, _, _, _, _, _, _).


% for propagation rule, can only remove symmetric heads if body goals cannot 
% affect computation (only catch "true" as such a body goal for now)

may_remove_symmetric(Body, Hs0, Hs, Info) :-
   simple_body(Body) ->
       remove_symmetric(Hs0, Hs), 
       (Hs0 == Hs -> % no removal
          check_double_samehead(Hs, Info)
        ; Info = nil
       ) 
     ; check_double_samehead(Hs0, Info),
       Hs0 = Hs.

simple_body(true) :- !.
simple_body((_G|Body)) :-
   simple_body(Body).

check_double_samehead([H1,H2], Info) :-
   functor(H1, F, A), functor(H2, F, A), !,
   Info = double(H1, H2).
check_double_samehead(_, nil).

 
remove_symmetric([H|Hs], [H|NewH]) :-
   remove_symmetric1(Hs, [H], NewH).

% check only first and second head only for now
remove_symmetric1([], _, []) :- !.
remove_symmetric1([H0|Hs0], Matches, Hs) :- 
   (is_symmetric(Matches, H0) ->
       remove_symmetric1(Hs0, Matches, Hs)
     ; remove_symmetric1(Hs0, [H0|Matches], Hs0),
       Hs = [H0|Hs0]
   ).


is_symmetric([H1|_], H2) :-
   \+ \+(H1 = H2),
   \+ \+is_pairsymmetric(H1,H2), !.
is_symmetric([_|Hs], H2) :-
   is_symmetric(Hs, H2).

is_pairsymmetric(H1,H2) :-
   copy_term((H2,H1), NPair),
   numbervars((H1,H2), 0, N),
   numbervars(NPair, 0, N),
   NPair == (H1,H2).

no_body((true|true)) :- !.
no_body(true).

single_simpogation([KH], [DH], Body, NBody, Type, Module) :-
   \+recorded('CHRdont_simpa_symmetric', Module),
   \+ \+(KH = DH),
   \+ \+is_pairsymmetric(KH, DH), !,
   (no_body(Body)  ->
      Type = nobody

    ; Type = twobodies,
      copy_term((KH,Body), (KH0,Body1)),
      copy_term((DH,Body), (DH0,Body2)),
      numbervars((KH0,Body1), 0, N),
      numbervars((DH0,Body2), 0, N),
      matchvars((KH0,Body1), (KH,Body), N, Vars1),
      matchvars((DH0,Body2), (DH,Body), N, Vars2),
      copy_term((Body,Vars1,Vars2),(NBody,Vars2,Vars1))
   ).
single_simpogation(_, _, _, _, Type, _M) :- Type = not_reducible.


matchvars(NumvarBody, Body, N, Vars) :-
  matchvars1(NumvarBody, Body, [], VarPairs),
  order_vars(N, VarPairs, Vars).

%matchvars1(Var,Var, VP, VP) :- var(Var), !. 
matchvars1('$VAR'(N), V, VP0, VP) :- !,
   (second_member(VP0, N, _) -> VP = VP0 ; VP = [V-N|VP0]).
matchvars1(Atom, Atom, VP, VP) :- atomic(Atom), !.
matchvars1(NumvarTerm, Term, VP0, VP) :-  % compound term
   functor(NumvarTerm, _Name, A),
   matchvars2(A, NumvarTerm, Term, VP0, VP).

matchvars2(0, _, _, VP, VP) :- !.
matchvars2(N, NumvarTerm, Term, VP0, VP) :-
   arg(N, NumvarTerm, NumvarSub),
   arg(N, Term, Sub),
   matchvars1(NumvarSub, Sub, VP0, VP1),
   N1 is N - 1,
   matchvars2(N1, NumvarTerm, Term, VP1, VP).

order_vars(0, [], []) :- !.
order_vars(N, VarPairs0, Vars) :-
   N1 is N - 1,
   % numbering starts from zero, so N1
   remove_secondmember(V-N1, VarPairs0, VarPairs1),
   Vars = [V|Vars0],
   order_vars(N1, VarPairs1, Vars0).

remove_secondmember(Obj-Item, [T-Index|L0], L) :-
   (Item == Index ->  L = L0, Obj = T
     ; L = [T-Index|L1],
       remove_secondmember(Obj-Item, L0, L1)
   ).


/*   shared_vars(H1, H2, Shared, _Rest),
   is_symmetric_with_shared(Shared, H1, H2).

is_symmetric_with_shared([], _, _) :- !.
is_symmetric_with_shared([Var|Vars], H1, H2) :-
   % need to check that all shared vars are in same position
   \+ \+(check_one_var(Var, H1, H2)),       
   is_symmetric_with_shared(Vars, H1, H2).
*/

check_one_var(*, H1, H2) :-
   copy_term(H1, H11), 
   variant(H11, H2).  % need copy_term to get rid of shared var

/* addto_processed(+ConsNo, +Rule, +Status, +Head, +KeepHeads, +DeleteHeads, 
       +Body, +Name, +PropagationNo, +Info, +Processed)
   adds information about rule Rule when the trying of the rule is initiated 
   by Head. ConsNo is the constraint number for Head. Status is either keep or 
   delete, indicating if Head is to be kept or deleted from the constraint 
   store if the rule is fired. KeepHeads and DeleteHeads are lists of the 
   other heads in rule (i.e. excluding Head) that are to be kept or deleted 
   after firing of rule. Body is the guard and body goals of the rule.
   Processed is the structure into which this information is to be added.
   PropagationNo is the propagation rule number for this propagation rule.
   Info is additional information that may be needed in transformation stage
*/
addto_processed(ConsNo, Rule, Status, Head, KeepHeads, DeleteHeads, Body, Name,
  NProp, Info, Processed) :-
     arg(ConsNo, Processed, RulesList),
     setarg(ConsNo, Processed, [rule(Rule,Status,Head,KeepHeads,DeleteHeads,Body, NProp, Info, Name)|RulesList]).



remove_head(1, [_Head|Heads], Heads) :- !.
remove_head(N, [Head|Heads0], [Head|Heads1]) :-
    N1 is N - 1,
    remove_head(N1, Heads0, Heads1).

construct_writeconstraint_code(SCons, Define, NNameRule) :-
    functor(SCons, F, A),
    atom_string(F, FS),
    append_strings(FS, "print", NNameS),
    atom_string(NName, NNameS), 

    arg(2, SCons, Constraint),
    NNameRule =.. [NName, SCons, Constraint],
    Define = (:- erase_macro(F/A), define_macro(F/A, NName/2, [write,goal])).

chr(File, Module) :-
  chr_clear,
  compile(File)@Module.












