% ----------------------------------------------------------------------
% 
% Hash table library
%
% System:	ECLiPSe Constraint Logic Programming System
% Copyright:	(C) Parc Technologies Ltd 1999
% Author/s:	Stefano Novello, IC-Parc
% Version:	$Id: hash.ecl,v 1.1 1999/08/05 16:28:02 js10 Exp $
%
% ----------------------------------------------------------------------

:- module_interface(hash).
:- export(hash_create/1).
:- export(hash_add/3).
:- export(hash_find/3).
:- export(hash_iter/2).
:- export(hash_next/4).
:- export(hash_last/1).
:- export(hash_list/3).
:- begin_module(hash).

:- local struct(hash_table(size,nb_elems,table)).
:- local struct(hash_elem(key,elem)).
:- local struct(hash_iter(next_index,bucket,table,eleft)).

hash_create(H) :-
	H = hash_table with [size:4,nb_elems:0,table:'[]'([],[],[],[])].

hash_add(H,Key,Elem) :-
	H = hash_table with [size:Size,nb_elems:Nb,table:T],
	hash(Key,Size,Index),
	Bucket is T[Index],
	Match = hash_elem with key:Key,
	( member(E,Bucket), E = Match ->
	    setarg(elem of hash_elem,E,Elem)
	;
	    Nb1 is Nb + 1,
	    MaxNb is Size * 3,
	    (Nb1 =< MaxNb ->
		Match = hash_elem with elem:Elem,
		setarg(nb_elems of hash_table,H,Nb1),
		setarg(Index,T,[Match|Bucket])
	    ;
		grow(H),
		hash_add(H,Key,Elem)
	    )
	).

hash_find(H,Key,Elem) :-
	H = hash_table with [size:Size,table:T],
	hash(Key,Size,Index),
	Bucket is T[Index],
	member(hash_elem with [key:Key,elem:E], Bucket),
	!,
	E=Elem.
	

grow(H):-
	H = hash_table with [size:OldSize,table:OldT],
	Size is 4 * OldSize,
	( for(I,1,Size) ,foreach([],Nulls) do true),
	T =.. [[]|Nulls],
	setarg(size of hash_table,H,Size),
	setarg(table of hash_table,H,T),
	( foreacharg(OldBucket,OldT),
	  param(T,Size)
	do
	    ( foreach(Elem,OldBucket),
	      param(T,Size)
	    do
		Elem = hash_elem with key:Key,
	    	hash(Key,Size,Index),
		arg(Index,T,Bucket),
		setarg(Index,T,[Elem|Bucket])
	    )
	).

	    
hash_last(hash_iter with [eleft:0]).

hash_iter(H,Iter) :-
	H = hash_table with [table:T,nb_elems:N],
	Bucket is T[1],
	Iter =  hash_iter with [next_index:2,bucket:Bucket,table:T,eleft:N].

hash_next(hash_iter with [next_index:S1,bucket:Bucket,table:T,eleft:N],Key,Elem,Iter) :-
	next_iter(N,S1,Bucket,T,Key,Elem,Iter).

next_iter(N,S,[],T,Key,Elem,Iter) :-
	N > 0,
	S1 is S+1,
	Bucket is T[S],
	next_iter(N,S1,Bucket,T,Key,Elem,Iter).
next_iter(N,S,[E|Bucket],T,Key,Elem,Iter) :-
	E = hash_elem with [key:Key,elem:Elem],
	N1 is N-1,
	Iter = hash_iter with [next_index:S,bucket:Bucket,table:T,eleft:N1].

hash_list(H,Keys,List) :-
	hash_iter(H,I),
	iter_list(I,Keys,List).

iter_list(hash_iter with [eleft:0],Keys,L) :-
	!,
	Keys=[],
	L=[].
iter_list(I,[K|Ks],[E|Es]) :-
	hash_next(I,K,E,I1),
	iter_list(I1,Ks,Es).

hash(Key,Size,Hash) :-
	atomic(Key),
	hash0(Key,Hash0),
	Hash1 is Hash0 /\ (Size - 1),
	Hash is Hash1 + 1.
hash([I|Is],Size,Hash) :-
	hash0(I,Hash0),
	hash0(Hash0,Is,Hash1),
	Hash2 is Hash1 /\ (Size - 1),
	Hash is Hash2 + 1.
	
hash0(Atom,Hash) :-
	atom(Atom),
	atom_string(Atom,String),
	hash0(String,Hash).
hash0(String,Hash) :-
	string(String),
	string_list(String,Clist),
	( foreach(C,Clist),
	  fromto(0,H0,H1,Hash)
	do
		H1 is H0 + 9*C
	).
hash0(Int,Hash) :-
	integer(Int),
	Hash = Int.

hash0(Hash0,[],Hash0).
hash0(Hash0,[K|Ks],Hash) :-
	hash0(K,Hash1),
	Hash2 is Hash0 + Hash1,
	hash0(Hash2,Ks,Hash).


	


