% ----------------------------------------------------------------------
% System:	ECLiPSe Constraint Logic Programming System
% Copyright (C) Imperial College London and ICL 1995-1999
% Version:	$Id: ifdef.pl,v 1.2 1999/08/07 15:50:20 js10 Exp $
% ----------------------------------------------------------------------

/*
 * SEPIA PROLOG SOURCE MODULE
 */

/*
 * sccsid("%W%		%E%").
 * sccscr("%Z%  Copyright 1989 ECRC GmbH ").
 */

/*
 * IDENTIFICATION:	ifdef.pl
 *
 * DESCRIPTION: 	This is a package that allows conditional compilation.
 *			The syntax is very similar to C:
 *
 *			:- ifdef Goal.
 *			...
 *			:- elseif Goal.
 *			...
 *			:- else.	% or :- else(Goal).
 *			...
 *			:- endif.	% or :- endif(Goal).
 *
 *			The conditionals can be nested.
 *
 */

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

:- module_interface(ifdef).

:- global (ifdef)/1.
:- op(1100, fy, [ifdef, elseif]).

:- begin_module(ifdef).

:- import
    compile_term_/2,
    compiled_stream/1,
    read_/3
   from sepia_kernel.

:- tool((ifdef)/1, ifdef_body/2).
:- make_local_array(err), define_error("ifdef without a matching endif", Err),
	setval(err, Err).

ifdef_body(Cond, Module) :-
	compiled_stream(S),
	(call(Cond, Module) ->
		then_part(S, Module)
	;
		else_part(S, Module)
	).

then_part(S, Module) :-
	read_to_next(S, List, End, yes, Module),
	compile_term_(List, Module),
	skip_to_endif(S, End, Module).

else_part(S, Module) :-
	read_to_next(S, _, End, no, Module),
	(End = elseif(Cond) ->
	    (call(Cond, Module) ->
		then_part(S, Module)
	    ;
		else_part(S, Module)
	    )
	;
	End \== endif ->
	    read_to_next(S, List, _, yes, Module),
	    compile_term_(List, Module)
	;
	    true
	).

skip_to_endif(_, endif, _) :- !.
skip_to_endif(S, _, Module) :-
	read_to_next(S, _, End, no, Module),
	skip_to_endif(S, End, Module).

read_to_next(S, List, End, True, Module) :-
    read_(S, Term, Module),
    (Term = (:- Goal) ->
	next_goal(S, Goal, List, End, True, Module)
    ;
    Term == end_of_file ->
	getval(err, Err),
	current_stream(File, _, S),
	error(Err, File)
    ;
	List = [Term|Next],
	read_to_next(S, Next, End, True, Module)
    ).

next_goal(_, elseif(Cond), [], elseif(Cond), _, _) :- !.
next_goal(_, else, [], else, _, _) :- !.
next_goal(_, else(_), [], else, _, _) :- !.
next_goal(_, endif, [], endif, _, _) :- !.
next_goal(_, endif(_), [], endif, _, _) :- !.
next_goal(S, Goal, List, End, True, Module) :-
    (True == yes -> call(Goal, Module); true),
    read_to_next(S, List, End, True, Module).

ifdef_error(N, File) :-
    (compiled_file(File, Line) ->
	true
    ;
	Line = '??'
    ),
    error_id(N, M),
    printf(error, "\n*** %s\n\tbefore line %w in the file %s\n",
	[M, Line, File]),
    flush(error),
    abort.

:- getval(err, Err), set_error_handler(Err, ifdef_error/2).
