% ----------------------------------------------------------------------
% System:	ECLiPSe Constraint Logic Programming System
% Copyright (C) Imperial College London and ICL 1997-1998
% Copyright (C) Parc Technologies Ltd 1999
% Version:	$Id: linearize.pl,v 1.3 1999/08/07 14:56:39 js10 Exp $
%
% Description:		Expression simplifier
%
% Authors:		J.Schimpf, IC-Parc
%
%
% A linear expression is normalised into a list (sum) of the form
%
%	[C0*1, C1*X1, C2*X2, ...]
%
% where Ci are numbers and Xi are distinct variables.
% The first (constant) term is always present, Ci (i>=1) are nonzero.
%
% linearize/3 converts a general arithmetic expression into a
% normalised linear form. The nonlinear parts are extracted using
% auxliliary variables, and returned as Residue, which is a list of
% Aux=NonLinExpr constraints.
%
% linearize/3 understands all arithmetic expressions plus
%
%	lin(Lin)	embedded normalised linear expression
%
%
% linrenorm/2 re-normalises a linear normal form expression after variable
% bindings, unifications.
%
% delinearize/2 converts back into normal arithmetic expression, but
% that may often not be needed since linearize/3 understands lin(Lin).
%
% TODO:
%	-- E1*E2-clause of linearize/7 always makes a choicepoint...
%	-- generic version of solver_map/4 for any constraint
%
% ----------------------------------------------------------------------


% ----------------------------------------------------------------------
:- module_interface(linearize).
% ----------------------------------------------------------------------

:- export
	linearize/3,
	linrenorm/2,
	delinearize/2.

% ----------------------------------------------------------------------
:- begin_module(linearize).
% ----------------------------------------------------------------------

linearize(Expr, Lin, Res) :-
	linearize(Expr, 0, C, L, [], Res, []),
	Lin0 = [C*1|L],
	linrenorm(Lin0, Lin).


% linearize(Expr, CstIn, CstOut, LinOut, LinIn, Res, ResT).
% Split arithmetic expression Expr into constant, linear and residue part.

:- mode linearize(?, +, -, -, +, -, +).
linearize(X, C, C, [1*X|L], L, R, R) :-
	var(X), !.
linearize(K, C0, C, L, L, R, R) :-
	number(K), !,
	C is C0+K.
linearize(subscript(Array,Index), C0, C, L, L0, R, R0) :-
	not nonground(Index), !,
	subscript(Array,Index,Elem),
	linearize(Elem, C0, C, L, L0, R, R0).
linearize(lin(Lin), C0, C, L, L0, R, R) :- !,
	Lin = [K*1|L1],
	C is C0+K,
	difflistify(L1, L, L0).
linearize(+E, C0, C, L, L0, R, R0) :- !,
	linearize(E, C0, C, L, L0, R, R0).
linearize(-E, C0, C, L, L0, R, R0) :- !,
	linearize((-1)*E, C0, C, L, L0, R, R0).
linearize(E1+E2, C0, C, L, L0, R, R0) :- !,
	linearize(E1, C0, C1, L1, L0, R1, R0),
	linearize(E2, C1, C, L, L1, R, R1).
linearize(sum(List), C0, C, L, L0, R, R0) :- !,
	linearize_sum(List, C0, C, L, L0, R, R0).
linearize(E1-E2, C0, C, L, L0, R, R0) :- !,
	linearize(E1, C0, C1, L1, L0, R1, R0),
	linearize((-1)*E2, C1, C, L, L1, R, R1).
linearize(E1*E2, C0, C, L, L0, R, R0) :- !,
	( make_prod_sum(E1, E2, ProdSum) ->
	    linearize_sum(ProdSum, C0, C, L, L0, R, R0)
	;
	    linearize(E1, 0, E1C, E1Lin, [], R1, R0),
	    linearize(E2, 0, E2C, E2Lin, [], R2, R1),
	    C is C0 + E1C*E2C,
	    lin_mult(E1C, E2Lin, L1, L0),
	    lin_mult(E2C, E1Lin, L2, L1),
	    ( E1Lin==[] ->
		L=L2, R=R2
	    ; E2Lin==[] ->
		L=L2, R=R2
	    ;
		delinearize(E1Lin,E1LinExpr),
		delinearize(E2Lin,E2LinExpr),
%		E1LinExpr = lin([0*1|E1Lin]),
%		E2LinExpr = lin([0*1|E2Lin]),
		L = [1*Aux|L2],
		% E1LinExpr*E2LinExpr could be simplified
		% and possibly a constant factored out.
		R = [Aux=E1LinExpr*E2LinExpr|R2]
	    )
	).
linearize(E, C, C, [1*Aux|L0], L0, [Aux=E|R0], R0) :-
	nonground(E), !.
linearize(E, C0, C, L, L, R, R) :-
	K is E, C is C0+K.	% ground(E)


    :- mode linearize_sum(?, +, -, -, +, -, +).
    linearize_sum(X, C, C, L, L, R, R) :-
    	var(X), !,
	error(4, _ is sum(X)).
    linearize_sum([], C, C, L, L, R, R) :- !.
    linearize_sum([E|Es], C0, C, L, L0, R, R0) :- !,
	linearize(E, C0, C1, L1, L0, R1, R0),
	linearize_sum(Es, C1, C, L, L1, R, R1).
    linearize_sum(X, C, C, L, L, R, R) :-
	error(5, _ is sum(X)).


    % lin_mult(+Cst, +LinIn, -LinOut, +LinOutT).
    lin_mult(K, CXs, CKXs, CKXs0) :-
	( K =:= 0 -> CKXs=CKXs0
	; lin_mult1(K, CXs, CKXs, CKXs0) ).

	:- mode lin_mult1(+,+,-,+).
	lin_mult1(_, [], CKXs, CKXs).
	lin_mult1(K, [C*X|CXs], [CK*X|CKXs1], CKXs) :-
	    CK is C*K,
	    lin_mult1(K, CXs, CKXs1, CKXs).

    make_prod_sum([], [], Sum) :- -?-> Sum=[].
    make_prod_sum([X|Xs], [Y|Ys], Sum) :- -?->
	Sum = [X*Y|XYs],
	make_prod_sum(Xs, Ys, XYs).


% ----------------------------------------------------------------------
% (re)normalise a linear term (see above) by collecting constants,
% collapsing multiple occurrences of variables and eliminating zeros.
% ----------------------------------------------------------------------

linrenorm(Lin, LinNorm) :-
	sort(2, >=, Lin, LinSorted),
	collapse_terms(LinSorted, LinNorm),
	( LinNorm=[Rhs*One|_], number(Rhs), One==1 -> true
	; writeln(error, "Error in normalise/2") ). 


collapse_terms([], []).
collapse_terms([X|Xs], R) :-
	collapse_terms(X, Xs, R).

collapse_terms(X, Xs, Res) :-
	X = F*V,
	( var(V), F =:= 0 ->
	    collapse_terms(Xs, Res)
	;
	    collapse_terms1(X, Xs, Res)
	).

collapse_terms1(Any, [], [Any]).
collapse_terms1(F1*V1, [F2*V2|More], Res) :-
	( number(V1), number(V2) ->
	    F is F1*V1+F2*V2,
	    collapse_terms(F*1, More, Res)
	; V1 == V2 ->
	    F is F1+F2,
	    collapse_terms(F*V1, More, Res)
	;
	    Res = [F1*V1|Res0],
	    collapse_terms(F2*V2, More, Res0)
	).


% ----------------------------------------------------------------------
% Normal form -> standard expressions
% ----------------------------------------------------------------------

delinearize(Norm, Expr) :-
	delinearize(Norm, 0, Expr).

delinearize([], Expr, Expr).
delinearize([F*X|Ts], Expr0, Expr) :-
	( Expr0 == 0 ->
	    Fact = F,   Expr1 = Term
	; sgn(F, -1) ->
	    Fact is -F, Expr1 = Expr0 - Term
	;
	    Fact = F,   Expr1 = Expr0 + Term
	),
	( nonvar(X) ->
	    Term = Fact
	; Fact =:= 1 ->
	    Term = X
	;
	    Term = Fact*X
	),
	delinearize(Ts, Expr1, Expr).


% ----------------------------------------------------------------------

difflistify([], DXs, DXs).
difflistify([X|Xs], [X|DXs], DXs0) :-
	difflistify(Xs, DXs, DXs0).
