% ----------------------------------------------------------------------
% System:	ECLiPSe Constraint Logic Programming System
% Copyright (C) Imperial College London and ICL 1995-1999
% Version:	$Id: lists.pl,v 1.4 1999/08/07 15:50:26 js10 Exp $
% ----------------------------------------------------------------------

/*
 * SEPIA PROLOG SOURCE MODULE
 *
 *	sccsid("%W%		%E%").
 *	sccscr("%Z%  Copyright 1989 ECRC GmbH ").
 *
 * IDENTIFICATION:	lists.pl 
 *
 * DESCRIPTION: 	List manipulation library predicates
 *			These are not protected and may be 
 *			redefined by user prolog if desired.
 *			No warning message is generated if 
 *			they are redefined.
 *
 * CONTENTS:
 *	member/2,
 *	memberchk/2,
 *	append/3,
 *	nonmember/2,
 *	delete/3,
 *	intersection/3,
 *	subtract/3,
 *	union/3,
 *	list_check/3,
 *	length/2.
 *
 */

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
:- module(lists).

% Make sure that the important operators are ok
:-	op(1100, xfy, ;),
	op(1050, xfy, ->).

:- global
	append/3,
	checklist/2,
	delete/3,
	flatten/2,
	intersection/3,
	length/2,
	maplist/3,
	member/2,
	memberchk/2,
	nonmember/2,
	print_list/1,
	reverse/2,
	subset/2,
	subtract/3,
	union/3.

:- export
	halve/3,
	splice/3.

:- tool(maplist/3, maplist_body/4).
:- tool(checklist/2, checklist_body/3).


% member/2
% (This version doesn't leave a choicepoint after the last result)
member(X, [H|T]) :- member(X, H, T).
member(X, X, _).
member(X, _, [H|T]) :- member(X, H, T).


memberchk(X,[X|_]) :- !.
memberchk(X,[_|T]):- memberchk(X,T).


append([],X,X).
append([X|L1],L2,[X|L3]):-
	append(L1,L2,L3).


nonmember(Arg,[Arg|_]) :-
	!,
	fail.
nonmember(Arg,[_|Tail]) :-
	!,
	nonmember(Arg,Tail).
nonmember(_,[]).


% delete (?Element, ?List, ?Result)
% Result is List with Element removed
delete(A, [A|C], C).
delete(A, [B|C], [B|D]) :-
	delete(A, C, D).

% intersection(L1, L2, L3)
% L3 is the intersection of L1 and L2, with arguments ordered as in L1

intersection([], _, []).
intersection([Head|L1tail], L2, L3) :-
	memberchk(Head, L2),
	!,
	L3 = [Head|L3tail],
	intersection(L1tail, L2, L3tail).
intersection([_|L1tail], L2, L3) :-
	intersection(L1tail, L2, L3).


% subtract(L1, L2, L3)
% L3 = L1 - L2

subtract([], _, []).
subtract([Head|L1tail], L2, L3) :-
	memberchk(Head, L2),
	!,
	subtract(L1tail, L2, L3).
subtract([Head|L1tail], L2, [Head|L3tail]) :-
	subtract(L1tail, L2, L3tail).


% union(L1, L2, L3)
% L3 is (L1-L2) + L2

union([], L, L).
union([Head|L1tail], L2, L3) :-
	memberchk(Head, L2),
	!,
	union(L1tail, L2, L3).
union([Head|L1tail], L2, [Head|L3tail]) :-
	union(L1tail, L2, L3tail).

% subset(Subset, S)

subset([],[]).
subset([X|L],[X|S]) :-
    subset(L,S).
subset(L, [_|S]) :-
    subset(L,S).


% list_check(List, 0, Length)
% succeeds iff List is a proper list and Length is its length

list_check(L, _, _) :-
	var(L),
	!,
	fail.
list_check([], N, N).
list_check([_|T], SoFar, Res) :-
	New is SoFar+1,
	list_check(T, New, Res).


% length(?List, ?Length)
% succeeds iff List is a list of length Length

length(List, Length) :-
	var(Length),
	!,
	length(List, 0, Length).
length(List, Length) :-
	integer(Length),
	Length >= 0,
	length1(List, Length).

:- mode length(?,+,?).
length([], Length, Length).
length([_|L], N, Length) :-
	N1 is N+1,
	length(L, N1, Length).

:- mode length1(?,+).
length1(L, 0) :- !, L=[].
length1([_|L], Length) :-
	N1 is Length-1,
	length1(L, N1).

maplist_body(_, [], [], _).
maplist_body(Pred, [H1|T1], [H2|T2], M) :-
	Pred =.. PL,
	append(PL, [H1, H2], NewPred),
	Call =.. NewPred,
	call(Call, M),
	maplist_body(Pred, T1, T2, M).

checklist_body(_, [], _).
checklist_body(Pred, [Head|Tail], M) :-
	Pred =.. PL,
	append(PL, [Head], NewPred),
	Call =.. NewPred,
	call(Call, M),
	checklist_body(Pred, Tail, M).

flatten(List, Flat) :-
	flatten(List, Flat, []).

flatten([], Res, Cont) :- -?-> !, Res = Cont.
flatten([Head|Tail], Res, Cont) :-
	-?->
	!,
	flatten(Head, Res, Cont1),
	flatten(Tail, Cont1, Cont).
flatten(Term, [Term|Cont], Cont).

reverse(List, Rev) :-
	reverse(List, Rev, []).

reverse([], L, L).
reverse([H|T], L, SoFar) :-
	reverse(T, L, [H|SoFar]).

halve(List, Front, Back) :-
	halve(List, List, Front, Back).

    halve([], Back0, Front, Back) :- !, Front=[], Back=Back0.
%   halve([_], Back0, Front, Back) :- !, Front=[], Back=Back0.	% Front=<Back
    halve([_], [X|Rs], Front, Back) :- !, Front=[X], Back=Rs.	% Front>=Back
    halve([_,_|Es], [X|Rs], [X|Fs], Back) :-
	halve(Es, Rs, Fs, Back).

splice([], Rs, LRs) :- !, LRs=Rs.
splice(Ls, [], LRs) :- !, LRs=Ls.
splice([L|Ls], [R|Rs], [L,R|LRs]) :-
	splice(Ls, Rs, LRs).


print_list([]).
print_list([H|T]) :-
	writeln(H),
	print_list(T).

:- skipped
	maplist/3,
	checklist/2,
	flatten/2,
	length/2,
	append/3,
	union/3,
	subtract/3,
	intersection/3,
	delete/3,
	nonmember/2,
	memberchk/2,
	member/2,
	reverse/2,
	print_list/1.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

