% ----------------------------------------------------------------------
% System:	ECLiPSe Constraint Logic Programming System
% Copyright (C) Imperial College London and ICL 1995-1999
% Version:	$Id: meter.pl,v 1.2 1999/08/07 15:50:29 js10 Exp $
% ----------------------------------------------------------------------

/*
 * SEPIA PROLOG DEMO LIBRARY MODULE
 *
 * sccsid('%W%             %E%').
 * sccscr('%Z%  Copyright 1990 ECRC GmbH ').
 *
 * IDENTIFICATION:	meter.pl
 *
 * AUTHOR:		Emmanuel van Rossum
 *
 * CONTENTS:             
 *			meter/0,
 *			close_meter/0.
 *
 * DESCRIPTION:
 *
 * this library contains a SEPIA demo showing the use of
 * graphics and events handlers to implement a graphical Prolog
 * preformance meter.
 *
 * USAGE: start it with meter/0 and close it with close_meter/0
 */


:- module(meter).

:- use_module(library(util)).

:- global
	meter/0,
	close_meter/0.

:- import
	get_flag_body/4
    from sepia_kernel.

:-
	make_local_array(flags),
	get_flag(variable_names, VN),
	get_flag(debug_compile, DC),
	setval(flags, flags(VN, DC)),
	nodbgcomp.

t_offset(55).
l_space(8).
scale(100).

window_data(0,0,220,150).

meter_data([
	      [scale(stack), 'Kb', 0],
	      [single_bar_max, 'global', 0, 0],
	      [single_bar_max, 'trail', 0, 0],
	      [single_bar_max, 'local', 0, 0],
	      [single_bar_max, 'control', 0, 0],
	      [integer, 'collect.', 0],
	      [scale(heap), 'Kb', 0],
	      [double_bar(heap), 'code heap', 0, 0],
	      [double_bar(heap), 'gen. heap', 0, 0],
	      [scale(cpu), 'sec', 0],
	      [double_bar(cpu), 'cpu time', 0, 0]
	  ]).

init_ref :-
	setval(heap_ref, 1024000),
	setval(cpu_ref, 10),
	Stack_ref is max(get_flag(global_size), get_flag(local_size)) * 1024,
	setval(stack_ref, Stack_ref),
	statistics(times, Times),
	setval(times, Times),
	statistics(gc_number, Collect),
	setval(collections, Collect).

meter :-
	global(window_data/4),
	open_2d(0,0,155, 100),
	set_colors,
	clear_view_surface(0),
	line_width(3),
	line_type(0),
	make_local_array(line(12, 3)),
	make_local_array(state),
	make_local_array(heap_ref),
	make_local_array(stack_ref),
	make_local_array(times),
	make_local_array(cpu_ref),
	meter_data(Data),
	init_data(Data),
	init_ref,
	setval(state, waiting),
	set_interrupt_handler(usr1, refresh_meter/0),
	set_interrupt_handler(alrm, update_meter/0),
	set_error_handler(154, start/3),
	set_error_handler(153, stop/3),
	set_error_handler(152, end_perf/2),
	refresh_meter,
	update_meter,
	set_timer(real, 0.5).

init_data(Lines) :-
	init_data(1, Lines).

init_data(_, []).
init_data(N, [[Type | [Text | Init]]|Lines]) :-
	setval(line(N, 0), Type),
	setval(line(N, 1), Text),
	setval(line(N, 2), Init),
	Next is N + 1,
	init_data(Next, Lines).

draw(Data) :-
	draw(1, Data).

draw(_, []).
draw(N, [Val | Data]) :-
	getval(line(N, 0), Type),
	update(N, Type, Val),
	Next is N + 1,
	draw(Next, Data).

update(N, single_bar, [Size]) :-
	get_ref(stack, Old_ref),
	(Size > Old_ref ->
	    make_new_ref(stack, Old_ref, Ref)
	;
	    Ref = Old_ref
	),
	Pc_size is (Size * scale) // Ref + 1,
	getval(line(N, 2), [Old_size]),
	setval(line(N, 2), [Pc_size]),
	update_line_from_to(N, Old_size, Pc_size).
update(N, single_bar_max, [Size]) :-
	get_ref(stack, Old_ref),
	(Size > Old_ref ->
	    make_new_ref(stack, Old_ref, Ref)
	;
	    Ref = Old_ref
	),
	Pc_size is (Size * scale) // Ref + 1,
	getval(line(N, 2), [Old_size, Old_max]),
	(Pc_size + 1 > Old_max ->
	    New_max is Pc_size + 1
	;
	    New_max is Old_max
	),
	setval(line(N, 2), [Pc_size, New_max]),
	update_max(N, Old_max, 0),
	update_line_from_to(N, Old_size, Pc_size),
	update_max(N, New_max, N).
update(N, double_bar(Type), [First, Second]) :-
	get_ref(Type, Old_ref),
	(Second > Old_ref ->
	    make_new_ref(Type, Old_ref, Ref)
	;
	    Ref = Old_ref
	),
	LinePos is l_space * N,
	Pc_first is fix((First * scale) / Ref),
	Pc_second is fix((Second * scale) / Ref),
	getval(line(N, 2), [Old_first, _]),
	setval(line(N, 2), [Pc_first, Pc_second]),
	First_pos is Pc_first + t_offset, Last_pos is Pc_second + t_offset,
	update_line_from_to(N, Old_first, Pc_first),
	line_color(N), line_width(1),
	line(First_pos, LinePos, Last_pos, LinePos),
	line_width(3).
update(N, scale(Type), _) :-
	get_scale(Type, Scale),
	getval(line(N, 2), [Old_scale]),
	setval(line(N, 2), [Scale]),
	Line_pos is N * l_space - 5,
	update_integer(80, Line_pos, N, Old_scale, Scale).
update(N, integer, [Int]) :-
	getval(line(N, 2), [Old_int]),
	setval(line(N, 2), [Int]),
	Line_pos is N * l_space - 2,
	t_offset(Col_pos),
	update_integer(Col_pos, Line_pos, N, Old_int, Int).

get_ref(heap, X) :-
	getval(heap_ref, X).
get_ref(cpu, X) :-
	getval(cpu_ref, X).
get_ref(stack, X) :-
	getval(stack_ref, X).

get_scale(heap, X) :-
	X is getval(heap_ref) // 1024.
get_scale(cpu, X) :-
	getval(cpu_ref, X).
get_scale(stack, X) :-
	X is getval(stack_ref) // 1024.

make_new_ref(stack, Old_ref, Ref) :-
	Factor = 2,
	Ref is Old_ref * Factor,
	setval(stack_ref, Ref),
	reset_line(1, Factor),
	reset_line(2, Factor),
	reset_line(3, Factor),
	reset_line(4, Factor),
	reset_line(5, Factor).
make_new_ref(heap, Old_ref, Ref) :-
	Factor = 2,
	Ref is Old_ref * Factor,
	setval(heap_ref, Ref),
	reset_line(7, Factor),
	reset_line(8, Factor),
	reset_line(9, Factor).
make_new_ref(cpu, Old_ref, Ref) :-
	Factor = 2,
	Ref is Old_ref * Factor,
	setval(cpu_ref, Ref),
	reset_line(11, Factor),
	reset_line(10, Factor).

reset_line(Line, Factor) :-
	getval(line(Line, 0), Type),
	getval(line(Line, 2), Val),
	rescale_data(Line, Type, Factor, Val, New_val),
	refresh_data(Line, Type, Val, 0),
	refresh_data(Line, Type, New_val, Line).

rescale_data(N, single_bar, Factor, [Val], [New_val]) :-
	New_val is Val // Factor,
	setval(line(N, 2), [New_val]).
rescale_data(N, single_bar_max, Factor, [Size, Max], [New_size, New_max]) :-
	New_size is Size // Factor,
	New_max is Max // Factor,
	setval(line(N, 2), [New_size, New_max]).
rescale_data(N, double_bar(_), Fact, [First, Second], [New_first, New_sec]) :-
	New_first is First // Fact,
	New_sec is Second // Fact,
	setval(line(N, 2), [New_first, New_sec]).
rescale_data(N, scale(_), Factor, [Old_scale], [New_scale]) :-
	New_scale is Old_scale * Factor,
	setval(line(N, 2), [New_scale]).

update_line_from_to(Line, Old, New) :-
	LinePos is l_space * Line,
	(New > Old ->
	    line_color(Line),
	    Start is Old + t_offset, End is New + t_offset,
	    line(Start, LinePos, End, LinePos)
	;
	(New < Old ->
	    Start is New + 1 + t_offset, End is t_offset + Old + 1,
	    line_color(0),
	    line(Start, LinePos, End, LinePos)
	;
	true
	)).

update_max(Line, Val, Color) :-
	Down is Line * l_space - 1,
	Up is Down + 2,
	Col_pos is t_offset + Val,
	line_width(1),
	line_color(Color),
	line(Col_pos, Down, Col_pos, Up),
	line_width(3).

update_integer(Pos, LinePos, Color, Old_int, Int) :-
	(Old_int = Int ->
	    true
	;
	(
	    text_color(0),
	    text(Pos, LinePos, Old_int),
	    text_color(Color),
	    text(Pos, LinePos, Int)
	)).

refresh_data :-
	between(1, 11, N),
	getval(line(N, 0), Type),
	getval(line(N, 2), Val),
	refresh_data(N, Type, Val, N),
	fail.
refresh_data.

refresh_data(N, scale(_), [Int], Color) :- !,
	LinePos is l_space * N - 5,
	text_color(Color),
	text(80, LinePos, Int).
refresh_data(N, integer, [Int], Color) :- !,
	LinePos is l_space * N - 2,
	Col_pos is t_offset,
	text_color(Color),
	text(Col_pos, LinePos, Int).
refresh_data(N, single_bar, [Size], Color) :-
	LinePos is l_space * N,
	Col_pos is t_offset,
	End is Col_pos + Size,
	line_color(Color),
	line(Col_pos, LinePos, End, LinePos).
refresh_data(N, single_bar_max, [Size, Max], Color) :-
	LinePos is l_space * N,
	Up is LinePos + 1,
	Down is LinePos - 1,
	Col_pos is t_offset,
	End is Col_pos + Size,
	Max_pos is Col_pos + Max,
	line_color(Color),
	line(Col_pos, LinePos, End, LinePos),
	line_width(1),
	line(Max_pos, Down, Max_pos, Up),
	line_width(3).
refresh_data(N, double_bar(_), [First, Second], Color) :-
	LinePos is l_space * N,
	Col_pos is t_offset,
	First_pos is First + t_offset,
	Second_pos is Second + t_offset,
	line_color(Color),
	line(Col_pos, LinePos, First_pos, LinePos),
	line_width(1),
	line(First_pos, LinePos, Second_pos, LinePos),
	line_width(3).

refresh_frame :-
	text_font_index(2),
	between(1, 11, N),
	getval(line(N, 0), Type),
	getval(line(N, 1), Text),
	text_color(N),
	refresh_frame(N, Type, Text),
	fail.
refresh_frame.

refresh_frame(N, scale(_), Text) :- !,
	Line_pos is N * l_space + 3, Up is Line_pos + 2, Down is Line_pos - 2,
	Text_pos is N * l_space - 5,
	Col_pos is t_offset,
	Last_col is Col_pos + scale,
	text(110, Text_pos, Text),
	line_color(N),
	line_width(1),
	line(Col_pos, Line_pos, Last_col, Line_pos),
	line(Col_pos, Up, Col_pos, Down),
	line(Last_col, Up, Last_col, Down),
	line_width(3).
refresh_frame(N, _, Text) :-
	Line_pos is N * l_space - 2,
	text(1, Line_pos, Text).

refresh_meter :-
	set_flag(enable_interrupts, off),
	refresh_frame,
	refresh_data,
	set_flag(enable_interrupts, on).

update_meter :-
	set_flag(enable_interrupts, off),
	(getval(state, waiting) ->
	    setval(state, refreshing),
	    statistics(global_stack_used, Global_used),
	    statistics(control_stack_used, Control_used),
	    statistics(local_stack_used, Local_used),
	    statistics(trail_stack_used, Trail_used),
	    statistics(code_heap_used, Code_used),
	    statistics(code_heap_allocated, Code_heap),
	    statistics(general_heap_used, General_used),
	    statistics(general_heap_allocated, General_heap),
	    Gc_number is statistics(gc_number) - getval(collections),
	    getval(times, [Old_user, Old_sys, Old_real]),
	    statistics(times, [User_time, Sys_time, Real_time]),
	    Laps_cpu is (User_time - Old_user) + (Sys_time - Old_sys),
	    Laps_real is Real_time - Old_real,
	    draw([
		     [],
		     [Global_used],
		     [Trail_used],
		     [Local_used],
		     [Control_used],
		     [Gc_number],
		     [],
		     [Code_used, Code_heap],
		     [General_used, General_heap],
		     [],
		     [Laps_cpu, Laps_real]
		 ]),
	    setval(state, waiting)
	;
	true
	),
	set_flag(enable_interrupts, on).

close_meter :-
	set_timer(real, 0),
	reset_interrupt_handler(usr1),
	reset_interrupt_handler(alrm),
	reset_error_handler(152),
	reset_error_handler(154),
	reset_error_handler(153),
	(current_array(state, _) ->
	    erase_array(state),
	    erase_array(heap_ref),
	    erase_array(times),
	    erase_array(cpu_ref),
	    erase_array(stack_ref),
	    erase_array(line/2)
	;
	    true
	),
	(get_flag_body(window_data/4, definition_module, meter, kegi) ->
	    local(window_data/4),
	    close_2d
	;
	    true
	).

end_perf(Err, N) :-
	close_meter,
	error(default(Err), N).

start(Err, G, M) :-
	init_ref,
	meter_data(Data),
	init_data(Data),
	clear_view_surface(0),
	refresh_meter,
	setval(state, waiting),
	update_meter,
	error(default(Err), G, M).

stop(N, X, Y) :-
	update_meter,
	setval(state, stopped),
	error(default(N), X, Y).

color(Number, Name) :-
	call(colour(Name, R, G, B), kegi),
	color_rgb(Number, R, G, B).

set_colors :-
	(getenv('COLOUR', _) ->
	    color(0, navy),
	    color(1, lightcyan),
	    color(7, lightcyan),
	    color(10, lightcyan),
	    color(2, yellow),
	    color(3, yellow),
	    color(4, yellow),
	    color(5, yellow),
	    color(6, seashell),
	    color(8, salmon),
	    color(9, pink),
	    color(11, green)
	;
	    true
	).

:-      getval(flags, flags(VN, DC)),
	set_flag(debug_compile, DC),
	set_flag(variable_names, VN),
	erase_array(flags).
