% ----------------------------------------------------------------------
% System:	ECLiPSe Constraint Logic Programming System
% Copyright (C) Imperial College London and ICL 1995-1998
% Copyright (C) Parc Technologies Ltd 1999
% Version:	$Id: mip.pl,v 1.6 1999/08/07 14:56:39 js10 Exp $
%
% MIP branch and bound in ECLiPSe using lib(eplex)
%
% J.Schimpf, IC-Parc, 1/96
% ----------------------------------------------------------------------

% ----------------------------------------------------------------------
:- module_interface(mip).
% ----------------------------------------------------------------------

:- use_module(library(eplex)).

:- export bb_inf/3.
:- export bb_inf_cplex/3.


% ----------------------------------------------------------------------
:- begin_module(mip).
% ----------------------------------------------------------------------

:- make_local_array(solver_prunes).
:- make_local_array(failures).

:- make_local_array(mip_bound).
:- make_local_array(mip_solution).


% ----------------------------------------------------------------------
% Variable selection
% ----------------------------------------------------------------------

integer_violation([X|Xs], FractVar, FractSol) :-
	lp_var_solution(X, Val),
	( abs(Val - round(Val)) >= int_tolerance ->
	    FractVar = X, FractSol = Val
	;
	    integer_violation(Xs, FractVar, FractSol)
	).

/*
integer_violation(Xs, FractVar, FractSol) :-
	integer_violation(Xs, _, 1.0, FractVar),
%	integer_violation(Xs, _, 0.0, FractVar),
	lp_var_solution(FractVar, FractSol),
%	call(get_var_index(FractVar, Idx),eplex),
%	Idx1 is Idx+1,
%	writeln(x(Idx1):FractSol),
	true.

integer_violation([], BestX, BestDiff, BestX) :-
	BestDiff >= int_tolerance,	% Did we actually find one?
	BestDiff < 1.0.
integer_violation([X|Xs], BestX, BestDiff, Res) :-
	lp_var_solution(X, Val),
	Diff is abs(Val - round(Val)),
%	( var(X), Val \== 0.0 ->
%	    call(get_var_index(X, Idx),eplex),
%	    Idx1 is Idx+1,
%	    write('    '),
%	    writeln(x(Idx1)=Val)
%	;
%	    true
%	),
	(
	    Diff >= int_tolerance,
%	    better(BestX, BestDiff, X, Diff)
%	    Diff < BestDiff
	    Diff =< BestDiff	% use 1.0 initially
%	    Diff > BestDiff	% use 0.0 initially
	->
	    integer_violation(Xs, X, Diff, Res)
	;
	    integer_violation(Xs, BestX, BestDiff, Res)
	).
*/

% prefer general integers to binaries
better(OldX, _OldDiff, _NewX, _NewDiff) :-
	free(OldX), !.
better(OldX, OldDiff, NewX, NewDiff) :-
	( var_range(OldX, _, 1.0) ->
	    ( var_range(NewX, _, 1.0) ->
		NewDiff =< OldDiff
	    ;
		true
	    )
	;
	    ( var_range(NewX, _, 1.0) ->
		fail
	    ;
		NewDiff =< OldDiff
	    )
	).


find_index([X|Xs], N, FractVar) :-
	( call(get_var_index(X, N),eplex) ->
	    FractVar=X
	;
	    find_index(Xs, N, FractVar)
	).


filter_integer_vars([], []).
filter_integer_vars([X|Xs], IntVars) :-
	var_type(X, Type),
	( Type == integer ->
	    IntVars = [X|IntVars0]
	;
	    IntVars = IntVars0
	),
	filter_integer_vars(Xs, IntVars0).

% ----------------------------------------------------------------------
% Branch-and-bound
% ----------------------------------------------------------------------

bb_inf_cplex(IVars, Expr, Inf) :-
	integers(IVars),	% to get proper typed_solution
	collect_lp_constraints_norm(Cstrs),
	lp_setup(Cstrs, min(Expr), [integers(IVars)], ProbHandle),
	lp_solve(ProbHandle, Inf),
	lp_get(ProbHandle, vars, V),
	lp_get(ProbHandle, typed_solution, S),
	V=S,
	lp_get(ProbHandle, statistics, [SS,SF|_]),
	writeln((solver_succs=SS, solver_fails=SF)),
	lp_cleanup(ProbHandle).


bb_inf(IVars, Expr, MinIpCost) :-
	setval(nodes,0),
	init_counters,

	lp_demon_setup(min(Expr), IpCost, [keep_basis(yes)], 9, [bounds], ProbHandle),

	lp_get(ProbHandle, vars, VArr),
	functor(VArr, _, NVars),
	VArr =.. [_|Vars],
	( var(IVars) -> filter_integer_vars(Vars, IVars) ; true ),
	length(IVars,NIVars),
	writeln(variables:NVars/intvars:NIVars),

	( integer_violation(IVars, BranchingVar, SplitValue) ->
	    ( 
		setval(mip_bound, 1e30),
		bb_loop(ProbHandle, IVars, BranchingVar, SplitValue, IpCost),
		var_range(IpCost, IpCostLwb, _),
		writeln('Better integer solution found, cost'=IpCostLwb),
		MipBound is IpCostLwb - int_tolerance,
		setval(mip_bound, MipBound),
		lp_get(ProbHandle, typed_solution, Solution),
		setval(mip_solution, Solution),
		fail				% continue
	    ;
		getval(mip_bound, MipBound),	% search finished
		MinIpCost is MipBound + int_tolerance,
		writeln('Integer optimum proven, cost'=MinIpCost),
		getval(mip_solution, Solution),
		% Binding the solution unfortunately triggers the solver again
		( VArr = Solution, true -> true
		; writeln(error, "Instantiating solution lead to failure") )
	    )
	;
	    var_range(IpCost, MinIpCost, _),
	    writeln('Fractional = integer solution found, cost'=MinIpCost),
	    lp_get(ProbHandle, typed_solution, Solution),
	    VArr=Solution
	),
	lp_get(ProbHandle, statistics, [SS,SF|_]),
	writeln((solver_succs=SS, solver_fails=SF)),
	lp_cleanup(ProbHandle).


bb_loop(ProbHandle, IVars, BranchingVar, SplitValue, IpCost) :-
	branch(BranchingVar, SplitValue, IpCost),
	( integer_violation(IVars, BranchingVar1, SplitValue1) ->
	    bb_loop(ProbHandle, IVars, BranchingVar1, SplitValue1, IpCost)
	;
	    true		% integer solution found
	).

branch(BranchingVar, SplitValue, IpCost) :-
	S is round(SplitValue),
	( S > SplitValue ->
	    ( lwb(BranchingVar, S)
	    ; reimpose_bound(IpCost), S1 is S-1, upb(BranchingVar, S1) )
	;
	    ( upb(BranchingVar, S)
	    ; reimpose_bound(IpCost), S1 is S+1, lwb(BranchingVar, S1) )
	),
	wake.	% <----- Solver is woken HERE -----

% The cost bound constraint must be re-imposed after every backtrack!
reimpose_bound(IpCost) :-
	getval(mip_bound, BestSoFar),
	upb(IpCost, BestSoFar).



init_counters :-
	setval(solver_succs, 0),
	setval(solver_fails, 0),
	setval(solver_prunes, 0),
	setval(failures, 0).
print_counters(Str) :-
	getval(failures, F),
	getval(solver_fails, SF),
	getval(solver_succs, SS),
	getval(solver_prunes, SP),
	writeln(Str, (failures=F, solver_succs=SS, solver_fails=SF, solver_prunes=SP)).

:- init_counters.


% ----------------------------------------------------------------------
% end_module(mip).
% ----------------------------------------------------------------------

