% ----------------------------------------------------------------------
% System:	ECLiPSe Constraint Logic Programming System
% Copyright (C) Imperial College London and ICL 1995-1999
% Version:	$Id: quintus.pl,v 1.6 1999/08/07 15:50:36 js10 Exp $
% ----------------------------------------------------------------------

/*
 * SEPIA PROLOG SOURCE MODULE
 */

/*
 * sccsid("%W%		%E%").
 * sccscr("%Z%  Copyright 1989 ECRC GmbH ").
 */

/*
 * IDENTIFICATION:	quintus.pl
 *
 * DESCRIPTION: 	Quintus prolog compatibility package
 *
 *
 * CONTENTS:     
 *
 */

:- module_interface(quintus).

:- use_module(library(cprolog)).
:- use_module(library(foreign)).
:- abolish log/2.
:- get_flag(instance/2, definition_module, M), M \== quintus ->
    abolish instance/2
    ; true.
:- export
	log/2,
	instance/2.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
%	SYNTAX
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

:-
    set_flag(syntax_option, blanks_in_nil),
    set_flag(syntax_option, '$VAR'),

    set_chtab(0'$, symbol),		% undo C-prolog
    set_chtab(0'|, special),

    call_explicit(op(0, fy, (not)), sepia_kernel),
    call_explicit(op(0, fy, (spied)), sepia_kernel),
    call_explicit(op(0, fx, (delay)), sepia_kernel),
    op(1100, xfy, (|)),
    op(1150, fx, [(meta_predicate), (multifile), (public), (mode), (dynamic)]),
    op(300, xfx, div),
    set_flag(macro_expansion, on).

:- import
	absolute_file_name/2
    from sepia_kernel.

:- export
	trans_bar/2,
	tr_lib/2,
	tr_q_module/2.

:-  define_macro((|)/2, trans_bar/2, []).
:-  define_macro((module)/2, tr_q_module/2, [local, top_only]).
:-  define_macro(library_directory/1, tr_lib/2, [local, clause]).

trans_bar(no_macro_expansion('|'(X,Y)), (X;Y)).

tr_q_module(no_macro_expansion(module(Module, ListOfPreds)),
	(module(Module), 
		use_module(library(quintus)),
		public(ListOfPreds))).

tr_lib(L, L) :-
	L = no_macro_expansion(library_directory(Path)),
	atom_string(Path, PathS),
	get_flag(library_path, P),
	(member(PathS, P) ->
	    true
	;
	    append(P, [PathS], New),
	    set_flag(library_path, New)
	).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

:- begin_module(quintus).
:- system.		% compiler directive to add the SYSTEM flag

:- import
	abolish_body/2,
	abolish_op_body/3,
	current_built_in_body/2,
	current_op_body/4,
	current_predicate_body/2,
	dbref_integer/2,
	dynamic_body/2,
	ensure_loaded/2,
	export_body/2,
	get_bip_error/1,
	get_flag_body/4,
	get_pager/1,
	global_body/2,
	global_op_body/4,
	import_body/2,
	list_predicates/2,
	more/1,
	nospy_body/2,
	printf_/8,
	retract_all_body/2,
	set_flag_body/4,
	tool_/2
   from sepia_kernel.

:- import
	fileerrors/0,
	nofileerrors/0,
	put/1
    from cprolog.

:- local
	(abolish)/1,
	display/1,
	(dynamic)/1,
	ensure_loaded/1,
	erase/1,
	op/3,
	put/2,
	recorded/3,
	recorda/3,
	recordz/3,
	use_module/1.

:-
	tool(retractall/1, retract_all_body/2),
	tool((abolish)/1, q_abolish_body/2),
	tool(incore/1, call/2),
	tool(format/2, format_body/3),
	tool(format/3, format_body/4),
	tool((multifile)/1, multifile_body/2),
	tool(nospyall/0, nospyall_body/1),
	tool(predicate_property/2, predicate_property_body/3),
	tool((public)/1, public_body/2),
	tool(op/3, op_body/4),
	tool((meta_predicate)/1, meta_predicate_body/2).

:-	set_stream(user_input, stdin),
	set_stream(user_output, stdout),
	set_stream(user_error, stderr).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% *** Loading Programs ***

no_style_check(X) :- var(X), !, fail.
no_style_check(single_var) :-
	get_flag(variable_names, check_singletons) ->
	    set_flag(variable_names, on)
	;
	    true.
no_style_check(discontiguous).
no_style_check(multiple).
no_style_check(all) :-
	no_style_check(single_var).

style_check(X) :- var(X), !, fail.
style_check(single_var) :- set_flag(variable_names, check_singletons).
style_check(discontiguous).
style_check(multiple).
style_check(all) :- 
	style_check(single_var).

multifile_body(Files, M) :-
	write('WARNING: declaring multifile procedure(s) as dynamic in '),
	writeln(multifile(Files)),
	qdynamic_body(Files, M).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% *** Online help ***

manual :-
	get_flag(installation_directory, Sepiadir),
	concat_string([Sepiadir, '/doc/user/'], ManDir),
	read_directory(ManDir, ' *User*', _, [File]),
	(get_pager(Pager) ->		% only if there is a paging program
		concat_string([Pager, ' "', ManDir, File, '"'], Comm),
		exec(Comm, [])
	;
		concat_strings(ManDir, File, Man),
		more(Man)
	).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% *** Control ***

otherwise.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% *** I/O ***

display(Term) :-
	call_explicit(display(stdout, Term), sepia_kernel).

format_body(Stream, List, ArgList, Module) :-
	List = [_|_],
	!,
	string_list(Format, List),
	format_body(Stream, Format, ArgList, Module).
format_body(Stream, Format, ArgList, Module) :-
	printf_(Stream, Format, ArgList, Module, 0'~, ErrF, ErrL, Res),
	(Res = 0 ->
		true
	;
		error(Res, format(Stream, ErrF, ErrL), Module)
	).

format_body(List, ArgList, Module) :-
	format_body(output, List, ArgList, Module).

open_null_stream(null).
flush_output(X) :- flush(X).
set_input(Stream) :- set_stream(input, Stream).
set_output(Stream) :- set_stream(output, Stream).
current_input(Stream) :- get_stream(input, Stream).
current_output(Stream) :- get_stream(output, Stream).

source_file(File) :-
	setof(X, P^(current_predicate(P), get_flag(P, source_file, X)), L),
	member(File, L).

source_file(Pred, File) :-
	(var(Pred) ->
	    current_predicate(F/A)
	;
	    true
	),
	functor(Pred, F, A),
	get_flag(F/A, source_file, File).

get0(S, T) :- get(S, T).

put(S, T) :-
	X is T,
	call_explicit(put(S, X), sepia_kernel).

character_count(Stream, N) :-
	at(Stream, N).

stream_position(S, N) :-
	at(S, N).

stream_position(Stream, Old, New) :-
	at(Stream, Old),
	seek(Stream, New).

ttyget0(X) :- get0(stdin, X).
ttyget(X) :- get(stdin, X).
ttyskip(X) :- skip(stdin, X).
ttyput(X) :- put(stdout, X).
ttynl :- nl(stdout).
ttytab(N) :- tab(stdout, N).
ttyflush :-
	current_stream(_, _, S),
	get_stream_info(S, device, tty),
	get_stream_info(S, mode, write),
	flush(S),
	fail;true.

portray_clause(Clause) :-
	writeclause(Clause).

line_count(Stream, Line1) :-
	get_stream_info(Stream, line, Line),
	Line1 is Line + 1.

op_body(_, _, [], _) :- !.
op_body(P, A, [O|L], M) :-
	!,
	op_body(P, A, O, M),
	op_body(P, A, L, M).
op_body(P, A, O, M) :-
	global_op_body(P, A, O, M),
	( current_op_body(P, A, O, M) ->
	    true
	;
	    abolish_op_body(O, A, M)	% remove the hiding local definition
	).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% *** Arithmetic ***

log(A, B) :- ln(A, B).			% should be in sicstus only, but then
integer(X,Y) :- T is X, Y is fix(T).	%  X is integer(log(2)) does not work
float(X) :- real(X).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% *** Term Conversion ***

numbervars('$VAR'(N), N, N1) :-
	!,
	N1 is N + 1.
numbervars(Term, N, Next) :-
	functor(Term, _, Arity),
	numbervars(0, Arity, Term, N, Next).

:- mode numbervars(++, ++, +, ++, -).
numbervars(Arity, Arity, _, Next, Next) :- !.
numbervars(I, Arity, Term, N, Next) :-
	I1 is I + 1,
        arg(I1, Term, Arg),
        numbervars(Arg, N, Mid),
        !,
        numbervars(I1, Arity, Term, Mid, Next).

atom_chars(Atom, List) :-
	var(Atom),
	!,
	string_list(String, List),
	atom_string(Atom, String).
atom_chars(Atom, List) :-
	atom(Atom),
	name(Atom, List).

number_chars(Number, List) :-
	var(Number),
	!,
	string_list(String, List),
	number_string(Number, String).
number_chars(Number, List) :-
	number_string(Number, String),
	string_list(String, List).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% *** Term Comparison ***

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% *** Environment ***

predicate_property_body(Pred, Property, M) :-
	var(Pred),
	!,
	(current_predicate_body(F/A, M); current_built_in_body(F/A, M)),
	functor(Pred, F, A),
	sepia_property(F/A, M, Property).
predicate_property_body(Pred, Property, M) :-
	functor(Pred, F, A),
	A < 256,
	sepia_property(F/A, M, Property).

sepia_property(P, M, compiled) :-
	get_flag_body(P, stability, static, M).
sepia_property(P, M, dynamic) :-
	get_flag_body(P, stability, dynamic, M).
sepia_property(P, M, built_in) :-
	get_flag_body(P, type, built_in, M).
sepia_property(P, M, exported) :-
	get_flag_body(P, visibility, V, M),
	(V == exported; V == global) -> true.
sepia_property(P, M, foreign) :-
	get_flag_body(P, call_type, external, M).
sepia_property(P, M, imported_from(Mi)) :-
	get_flag_body(P, visibility, imported, M),
	get_flag_body(P, definition_module, Mi, M).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% *** Debugging ***

nospyall_body(M) :-
	nospy_body(_, M),
	writeln('All spypoints removed').

unknown(Old, New) :-
	(nonvar(Old) ->
	    error(5, unknown(Old, New))
	;
	    (get_error_handler(68, fail/0, _) ->
		Old = fail
	    ;
		Old = trace
	    ),
	    (New == trace ->
		reset_error_handler(68)
	    ;
	    New == fail ->
		set_error_handler(68, fail/0)
	    ;
		error(5, unknown(Old, New))
	    )
	).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% *** Modules ***

public_body([], _) :- !.
public_body([Proc|Procs], M) :-
	!,
	globalize([Proc|Procs], M).
public_body(Procs, M) :-
	export_body(Procs, M).

current_module(Module, File) :-
	setof((M,F), P^(current_predicate(P),
	    get_flag(P, source_file, F),
	    get_flag(P, definition_module, M)), L),
	member((Module,File), L).

:- system_debug.
%:- system.
globalize([], _).
globalize([Pred|Rest], M) :-
	export_body(Pred, M),
	globalize(Rest, M).

:- tool(use_module/2, use_module_body/3).
:- tool(use_module/1, use_module_body/2).
:- tool(ensure_loaded/1, use_module_body/2).

use_module_body(F, L, M) :-
	ensure_loaded(F, M),
	(F = library(FM) ->
		true
	;
		FM = F
	),
	import_list(L, FM, M).

use_module_body(F, M) :-
	ensure_loaded(F, M),
	import_list(F, M).

import_list([], _, _).
import_list([Pred|L], F, M) :-
	import_body((Pred from F), M),
	import_list(L, F, M).

import_list([], _).
import_list([File|L], M) :-
	!,
	import1(File, M),
	import_list(L, M).
import_list(File, M) :-
	import1(File, M).

import1(library(File), M) :-
	!,
	import1(File, M).
import1(File, M) :-
	pathname(File, _, ModS),
	atom_string(Module, ModS),
	current_module(Module),
	!,
	import_body(Module, M).
import1(_, _).

meta_predicate_body((A,B), M) :-
	!,
	meta_predicate_body(A, M),
	meta_predicate_body(B, M).
meta_predicate_body(Def, M) :-
	functor(Def, N, A),
	tool_(N/A, M),
	printf(error, '*** Warning: the meta_predicate definition %w%n    must be manually replaced by a tool definition%n%b', [Def]).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% *** Dynamic Database ***

q_abolish_body(Name/Arity, Module) :-
	!,
	abolish_body(Name/Arity, Module).
q_abolish_body(:(Module, Spec), _) :-
	!,
	q_abolish_body(Spec, Module).
q_abolish_body([], _) :- !.		% wrong, but compatible
q_abolish_body([Spec|T], Module) :-
	!,
	q_abolish_body(Spec, Module),
	q_abolish_body(T, Module).
q_abolish_body(Atom, Module) :-
	atom(Atom),			% this is wrong in Q2.0!
	!,
	(list_predicates(Atom/N, Module) ->
		abolish_body(Atom/N, Module),
		fail
	;
		true
	).
q_abolish_body(Term, _) :-
	error(5, abolish(Term)).

:- tool((abolish)/2, abolish_body/3).
abolish_body(Name, Arity, Module):-
	q_abolish_body(Name/Arity, Module).

:- tool((dynamic)/1, qdynamic_body/2).

qdynamic_body(M:P, _) :-
    dynamic_body(P, M).
qdynamic_body(F/A, M) :-
    dynamic_body(F/A, M).
qdynamic_body((P1, P2), M) :-
    qdynamic_body(P1, M),
    qdynamic_body(P2, M).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% *** Internal Database ***

current_key(KeyName, KeyTerm) :-
	current_record(IKey),
	external_key(IKey, KeyTerm),
	functor(KeyTerm, KeyName, _).

external_key(Key, KeyN) :-
	atom(Key),
	atom_string(Key, KeyS),
	char_int(NS, 0),		% can't write "\000" in this syntax
	string_length(KeyS, N),
	substring(KeyS, NS, N),
	!,
	term_string(KeyN, KeyS).
external_key(Key, Key).


internal_key(Key, IKey) :-
	var(Key), !,
	current_record(IKey),
	external_key(IKey, Key).
internal_key(Key, IKey) :-
	number(Key), !,
	char_int(NS, 0),		% can't write "\000" in this syntax
	concat_atom([Key, NS], IKey).
internal_key(Key, Key).

% Possible modes: recorded(+,-,-) recorded(-,-,+) recorded(-,-,-)
recorded(Key, Term, QRef) :-
	var(QRef), !,
	internal_key(Key, IKey),
	call_explicit(recorded(IKey, Term, DbRef), sepia_kernel),
	dbref_integer(DbRef, Ref),
	QRef = '$ref'(Ref, 0).
recorded(Key, Term, '$ref'(Ref, 0)) :-
	internal_key(Key, IKey),
	call_explicit(recorded(IKey, Term, DbRef), sepia_kernel),
	dbref_integer(DbRef, Ref),
	!.

recorda(Key, Term, '$ref'(Ref, 0)) :-
	nonvar(Key),
	internal_key(Key, NewKey),
	call_explicit(recorda(NewKey, Term, DbRef), sepia_kernel),
	dbref_integer(DbRef, Ref).

recordz(Key, Term, '$ref'(Ref, 0)) :-
	nonvar(Key),
	internal_key(Key, NewKey),
	call_explicit(recordz(NewKey, Term, DbRef), sepia_kernel),
	dbref_integer(DbRef, Ref).

erase('$ref'(Ref, 0)) :-
	dbref_integer(DbRef, Ref),
	call_explicit(erase(DbRef), sepia_kernel).

instance('$ref'(Ref, 0), Term) :-
	dbref_integer(DbRef, Ref),
	referenced_record(DbRef, Term).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% *** Grammar Rules ***

% only temporary
expand_term(A, B) :-
	term_expansion(A, B),
	!.
expand_term(A, A).

% To avoid calling an undefined procedure.
term_expansion(_, _) :- fail.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% *** Miscellaneous ***

prolog_flag(Flag, Old, New) :-
	var(Flag), 
	!,
	error(4, prolog_flag(Flag, Old, New)).
prolog_flag(Flag, Old, New) :-
	var(New), 
	Old \== New,
	!,
	error(4, prolog_flag(Flag, Old, New)).
prolog_flag(Flag, Old, New) :-
	prolog_flag(Flag, Old),
	set_quintus_flag(Flag, New).

set_quintus_flag(character_escapes, New) :-
	(New == on ->
	    set_chtab(0'\, escape)
	;
	New == off ->
	    set_chtab(0'\, symbol)
	).
set_quintus_flag(debugging, New) :-
	(New == off ->
		set_flag(debugging, nodebug)
	;
		set_flag(debugging, New)
	).
set_quintus_flag(fileerrors, New) :-
	(New == on ->
		fileerrors
	;
	New == off ->
		nofileerrors
	).
set_quintus_flag(single_var_warnings, New) :-
	(New == on ->
	    set_flag(variable_names, check_singletons)
	;
	get_flag(variable_names, off) ->
	    true
	;
	    set_flag(variable_names, on)
	).
set_quintus_flag(unknown, New) :-
	unknown(_, New).
set_quintus_flag(gc, New) :-
	set_flag(gc, New).
set_quintus_flag(gc_margin, New) :-
	set_flag(gc_interval, New).
set_quintus_flag(gc_trace, New) :-
	(New == on ->
		set_flag(gc, verbose)
	;
	New == off ->
		set_flag(gc, on)
	).

prolog_flag(character_escapes, Old) :-
	(
		get_chtab(0'\, escape) -> Old = on ; Old = off
	).
prolog_flag(debugging, Old) :-
	get_flag(debugging, Mode),
	(Mode == nodebug ->
		Old = off
	;
	Mode == leap ->
		Old = debug
	;
		Old = trace
	).
prolog_flag(fileerrors, Old) :-
	(get_error_handler(170, nofileerrors_handler/2, _) ->
		Old = off
	;
		Old = on
	).
prolog_flag(single_var_warnings, Old) :-
	(get_flag(variable_names, check_singletons) ->
	    Old = on
	;
	    Old = off
	).
prolog_flag(unknown, Old) :-
	unknown(Old, Old).
prolog_flag(gc, Old) :-
	get_flag(gc, OldF),
	(OldF == off -> Old = off; Old = on).
prolog_flag(gc_margin, Old) :-
	get_flag(gc_interval, Old).
prolog_flag(gc_trace, Old) :-
	(get_flag(gc, verbose) ->
		Old = on
	;
		Old = off
	).
prolog_flag(module, M) :-
	get_flag(toplevel_module, M).

version.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% *** System Dependent ***

is_lower(X):- X < 128, X > 96.
is_upper(X):- X > 64, X < 91.
is_digit(X):- X > 47, X < 58.

unix(cd(Path)) :-
	cd(Path).
unix(cd) :-
	getenv('HOME', Home),
	cd(Home).
unix(shell(X)) :-
	getenv('SHELL', Shell),
	concat_string([Shell, ' -c "', X, '"'], Command),
	sh(Command).
unix(system(X)) :-
	sh(X).
unix(system(X, Status)) :-
	getenv('SHELL', Shell),
	concat_string([Shell, ' -c "', X, '"'], Command),
	exec(sh, [S], Pid),
	printf(S, '%w\nexit\n%b', Command),
	close(S),
	wait(Pid, Status).
unix(shell) :-
	getenv('SHELL', Shell),
	sh(Shell).
unix(argv(L)) :-
	argc(N),
	args_list(N, [], L).
unix(exit(N)) :-
	exit(N).

:- mode args_list(++, +, -).
args_list(1, L, L) :- !.
args_list(N, L, M) :-
	N1 is N - 1,
	argv(N1, S),
	(number_string(A, S) -> true ; atom_string(A, S)),
	args_list(N1, [A|L], M).

% put_line(list of charcters)
% writes the list of characters and a newline to the current output 

put_line([]) :-
	!,
	nl.
put_line([H|T]) :-
	put(H),
	put_line(T).

gc :-	set_flag(gc, on).
nogc :-	set_flag(gc, off).

stream_code(S, C) :-
	get_stream(S, C).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

:- Preds = (
	    (abolish)/1,
	    (abolish)/2,
	    atom_chars/2,
	    character_count/2,
	    current_input/1,
	    current_output/1,
	    current_key/2,
	    current_module/2,
	    display/1,
	    (dynamic)/1,
	    ensure_loaded/1,
	    erase/1,
	    expand_term/2,
	    float/1,
	    flush_output/1,
	    format/2,
	    format/3,
	    gc/0,
	    get0/2,
	    incore/1,
	    integer/2,
	    is_digit/1,
	    is_lower/1,
	    is_upper/1,
	    line_count/2,
	    manual/0,
	    (meta_predicate)/1,
	    (multifile)/1,
	    no_style_check/1,
	    nogc/0,
	    nospyall/0,
	    number_chars/2,
	    numbervars/3,
	    op/3,
	    open_null_stream/1,
	    otherwise/0,
	    portray_clause/1,
	    predicate_property/2,
	    prolog_flag/2,
	    prolog_flag/3,
	    (public)/1,
	    put/2,
	    put_line/1,
	    recorda/3,
	    recorded/3,
	    recordz/3,
	    retractall/1,
	    set_input/1,
	    set_output/1,
	    source_file/1,
	    source_file/2,
	    stream_code/2,
	    stream_position/2,
	    stream_position/3,
	    style_check/1,
	    term_expansion/2,
	    ttyflush/0,
	    ttyget/1,
	    ttyget0/1,
	    ttynl/0,
	    ttyput/1,
	    ttyskip/1,
	    ttytab/1,
	    unix/1,
	    unknown/2,
	    use_module/1,
	    use_module/2,
	    version/0
	),
	skipped(Preds),
	(export Preds),
	skipped((instance/2, log/2)).

:- (current_module(user) ->
	true
    ;
	create_module(user),
	call(use_module(quintus), user)
    ).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
%	MODULE INITIALIZATION
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
