% ----------------------------------------------------------------------
% System:	ECLiPSe Constraint Logic Programming System
% Copyright (C) Imperial College London and ICL 1995-1999
% Version:	$Id: suspend.pl,v 1.7 1999/08/07 15:50:42 js10 Exp $
% ----------------------------------------------------------------------

%
% SEPIA PROLOG MODULE
%
% sccsid("@(#)suspend.pl	1.18             95/12/20").
% sccscr("@(#)  Copyright 1992 ECRC GmbH ").
%
% IDENTIFICATION:	suspend.pl
%
% AUTHOR:		Micha Meier
%			Joachim Schimpf
%
% CONTENTS:		Coroutining with three suspension lists:
%
%			- one for goals to be woken on instantiation only
%			- one for binding with any metaterm
%			- one for general "constrained-ness"
%
%			The code for metaterm handlers should be used
%			as template for writing user-defined extensions;
%			simply replace 'suspend' in the predicate names by
%			the name of the attribute and re-use the code.
%

:- module_interface(suspend).

:- export tr_if_suspend/3.

% This particular structure is already predefined in the kernel
%:- use_module(library(structures)).
%:- define_struct(suspend(inst, constrained, bound)).


:- begin_module(suspend).

:- import
	get_flag_body/4,
	insert_suspension/4,
	setarg/3,
	suspensions_to_goals/3
    from sepia_kernel.

%----------------------------------------------------------------
% attribute declaration
%----------------------------------------------------------------

:- meta_attribute(suspend, [
	unify:			unify_suspend/2,
	compare_instances:	compare_instances_suspend/3,
	delayed_goals:		delayed_goals_suspend/3,
	delayed_goals_number:	delayed_goals_number_suspend/2
	]).

%----------------------------------------------------------------
% unify
%----------------------------------------------------------------

:- pragma(system).
:- pragma(nodebug).

% unify_suspend(+Term, Attribute)
unify_suspend(_, Attr) :-
    /*** ANY + VAR ***/
    var(Attr).			% Ignore if no attribute for this extension
unify_suspend(Term, Attr) :-
    compound(Attr),
    unify_term_suspend(Term, Attr).


% We wake every time a variable is touched.
:- mode unify_term_suspend(?, +).
unify_term_suspend(Term, AttrX) :-
    atomic(Term),		% The metaterm was instantiated, wake all
    /*** NONVAR + META ***/
    schedule_suspensions(constrained of suspend, AttrX),
    AttrX = suspend with [bound:B, inst:I],
    schedule_woken(B),
    schedule_woken(I).
unify_term_suspend(Term, AttrX) :-
    compound(Term),		% The metaterm was instantiated, wake all
    /*** NONVAR + META ***/
    schedule_suspensions(constrained of suspend, AttrX),
    AttrX = suspend with [bound:B, inst:I, constrained:Cnew],
    schedule_woken(B),
    schedule_woken(I),
    % the constrained list may still contain demons after scheduling,
    % forward them to the variables in the bound term
    ( Cnew = [] ->
    	true
    ;
	term_variables(Term, Vars),
	( Vars = [] -> true ; forward_constrained_list(Cnew, Vars) )
    ).
unify_term_suspend(Y{AttrY}, AttrX) :-
    -?->
    unify_suspend_suspend(Y, AttrX, AttrY).

unify_suspend_suspend(_, AttrX, AttrY) :-
    var(AttrY),			% No attribute for this extension
    /*** VAR + META ***/
    AttrX = AttrY.		% Keep both lists, do not wake
unify_suspend_suspend(_, AttrX, AttrY) :-
    nonvar(AttrY),
    /*** META + META ***/
    AttrX = suspend with [inst:XI-YI, bound:XB],
    AttrY = suspend with [inst:YI-YI0, bound:YB],
    setarg(inst of suspend, AttrY, XI-YI0),
    setarg(bound of suspend, AttrY, []),
    merge_suspension_lists(constrained of suspend, AttrX, constrained of suspend, AttrY),
    schedule_suspensions(constrained of suspend, AttrY),
    schedule_woken(XB),
    schedule_woken(YB).

forward_constrained_list([], _Vars).
forward_constrained_list([Susp|Susps], Vars) :-
    ( is_suspension(Susp) ->
	insert_suspension(Vars, Susp, constrained of suspend)
    ;
	true	% dead suspension
    ),
    forward_constrained_list(Susps, Vars).


%----------------------------------------------------------------
% compare_instances
%----------------------------------------------------------------

compare_instances_suspend(Res, T1, T2) :-
    compare_instances(Res, T1, T2).

%----------------------------------------------------------------
% delayed goals
%----------------------------------------------------------------

delayed_goals_suspend(_{Attr}, Goals, G0) :-
    -?->
    dg_suspend(Attr, Goals, G0).

dg_suspend(AttrX, Goals, Goals) :-
    /*** VAR ***/
    var(AttrX),
    !.
dg_suspend(suspend with [inst:I-_, bound:B, constrained:C], Goals, G0) :-
    /*** META ***/
    suspensions_to_goals(I, Goals, G1),
    suspensions_to_goals(B, G1, G2),
    suspensions_to_goals(C, G2, G0).

%----------------------------------------------------------------
% delayed goals number
%----------------------------------------------------------------

delayed_goals_number_suspend(_{AttrX}, N) :-
    -?->
    dgn_suspend(AttrX, N).

dgn_suspend(AttrX, 0) :-
    /*** VAR ***/
    var(AttrX),
    !.
dgn_suspend(suspend with [inst:I-_, bound:B, constrained:C], N) :-
    /*** META ***/
    count_active_suspensions(I, 0, N0),
    count_active_suspensions(B, N0, N1),
    count_active_suspensions(C, N1, N).

count_active_suspensions(Susps, N0, N) :-
    var(Susps),
    !,
    N = N0.
count_active_suspensions([], N, N).
count_active_suspensions([Susp|Susps], N0, N) :-
    ( suspension_to_goal(Susp, _, _) ->
	    N1 is N0 + 1
    ;
	    N1 = N0
    ),
    count_active_suspensions(Susps, N1, N).


:- untraceable
	unify_suspend/2,
	compare_instances_suspend/3,
	delayed_goals_suspend/3,
	delayed_goals_number_suspend/2.


%----------------------------------------------------------------
% Sepia's delay clauses (backward compatibility)
% Source transformation for if/2
%----------------------------------------------------------------

:- define_macro((if)/2, tr_if_suspend/3, [clause, global]).

:- mode tr_if_suspend(+, -, +).
tr_if_suspend(Clause, (Head :- -?-> NewBody, Susp), M) :-
	Clause = no_macro_expansion(delay Head if Body),
	translate_goal(Body, NewBody, VarList, [], BVarList, []),
	check_varlists(VarList, BVarList, Clause),
	strip_varlist(VarList, VarList1),
	strip_varlist(BVarList, BVarList1),
	handle_suspension(Head, VarList1, BVarList1, Susp, M).

:- mode translate_goal(?, -, ?, -, ?, -).
translate_goal(Goal,	Goal,	Vars,	Vars, BV, BV) :- var(Goal), !.
translate_goal((Goal, Goals), (NewGoal, NewGoals), Out, In, BOut, BIn) :- !,
	translate_goal(Goal, NewGoal, Mid, In, BMid, BIn),
	translate_goal(Goals, NewGoals, Out, Mid, BOut, BMid).
translate_goal((Goal -> Goals), (NewGoal -> NewGoals), Out, In, BOut, BIn) :- !,
	translate_goal(Goal, NewGoal, Mid, In, BMid, BIn),
	translate_goal(Goals, NewGoals, Out, Mid, BOut, BMid).
translate_goal((Goal ; Goals), (NewGoal, Out=VarsL, BOut = BVarsL ;
				NewGoals, Out=VarsR, BOut = BVarsR),
		[Out], In, [BOut], BIn) :- !,
	translate_goal(Goal, NewGoal, VarsL0, In, BVarsL0, BIn),
	check_varlists(VarsL0, BVarsL0, Goal),
	strip_varlist(VarsL0, VarsL),
	strip_varlist(BVarsL0, BVarsL),
	translate_goal(Goals, NewGoals, VarsR0, In, BVarsR0, BIn),
	check_varlists(VarsR0, BVarsR0, Goal),
	strip_varlist(VarsR0, VarsR),
	strip_varlist(BVarsR0, BVarsR).
translate_goal(var(X),		var(X),			[X|Out],    Out, V,V) :-
	!.
translate_goal(X \== Y,
		(call_explicit(\==(X, Y, Vars), sepia_kernel),
		    ( Vars = [Var] -> Out = [Var|In], BOut = BIn
		    ; BOut = [Vars|BIn], Out = In
		    )
		),
		[Out], In, [BOut], BIn) :- !.
translate_goal(nonground(X),	nonground(X, Var),	[Var|Out],  Out, V,V) :-
	!.
translate_goal(nonground(N, X),	nonground(N, X, Vars),	[Vars|Out], Out, V,V) :-
	!.
translate_goal(Goal,	Goal, 	Vars,	Vars, BVars, BVars).

handle_suspension(Head, List, BL,
    (!, make_suspension(Head, Prio, S), Body), M) :-
	handle_i_susp(S, List, BL, Body),
	functor(Head, F, A),
	(get_flag_body(F/A, waking, lazy, M) ->
	    Prio = 3
	;
	    Prio = 2
	).

handle_i_susp(Susp, [], BList, Body) :-
	-?->
	!,
	bound_suspension(Susp, BList, Body).
handle_i_susp(Susp, List, [], Body) :-
	-?->
	!,
	inst_suspension(Susp, List, Body).
handle_i_susp(Susp, List, BList, (BI, BB)) :-
	inst_suspension(Susp, List, BI),
	bound_suspension(Susp, BList, BB).

inst_suspension(S, List,
	insert_suspension(List, S, inst of suspend, suspend)).

bound_suspension(S, List,
	insert_suspension(List, S, bound of suspend, suspend)).


:- mode check_varlists(+,+,?).
check_varlists([], [], Culprit) :- !,
	error(272, Culprit).
check_varlists(_, _, _).

:- mode strip_varlist(+,-).
strip_varlist([], []) :- !.
strip_varlist([X], X) :- !.	% avoid list if single element
strip_varlist(L, L).

