% ----------------------------------------------------------------------
% System:	ECLiPSe Constraint Logic Programming System
% Copyright (C) Imperial College London and ICL 1995-1999
% Version:	$Id: toplevel.pl,v 1.15 1999/08/07 15:50:45 js10 Exp $
% ----------------------------------------------------------------------

%
% $Id: toplevel.pl,v 1.15 1999/08/07 15:50:45 js10 Exp $
%
% This is a toplevel which communicates with a user interface via
% 4 queue-streams and a string-based protocol.
%
%
% The queue-stream toplevel_in accepts the following command syntax
% in exdr-format:
%
% call(GoalString)
%	Execute the specified goal under control of the toplevel.
%
%	The result will be returned on the answer_out stream as a
%	string in <exdr>-format, and is one of:
%
%		"Yes"
%		"More"
%		"No"
%		"Abort"
%
%	Variable bindings will be printed (in no particular format)
%	on the toplevel_out stream.
%
% more
%	Request backtracking into the last goal (which must have
%	succeeded with a "More" result). The result is as for call(...).
%
% end
%	Cut and fail the last goal (which has succeeded with a "More"
%	or "Yes" result).
%	
% exit
%	Terminate the toplevel loop.
%

%----------------------------------------------------------------------
:- module_interface(toplevel).
%----------------------------------------------------------------------
:- global gui_toplevel/0.

:- lib(development_support).
:- lib(tracer_tcl).

%----------------------------------------------------------------------
:- begin_module(toplevel).
%----------------------------------------------------------------------

:- import
	get_cut/1,
	compiled_file_handler/3,
	set_default_error_handler/2,
	tr_goals/3,
	top_call_local/2
   from sepia_kernel.

:- open(queue(""), read, toplevel_in).
:- open(queue(""), write, toplevel_out).
:- open(queue(""), write, answer_out, [yield(on)]).


gui_toplevel :-
	set_default_error_handler(139, compiled_file_handler/3),
	reset_error_handler(139),

	write_exdr(toplevel_out, "Yes"),
	repeat,
	    block(gui_toplevel1, Tag, top_abort(Tag)),
	!.

    top_abort(end) ?- !.
    top_abort(abort) ?- !,
	drain_stream(toplevel_in),
	writeln(answer_out, "Abort."),
	flush(answer_out),
    	write_exdr(toplevel_out, "Abort"),
	fail.
    top_abort(Tag) :-
	block(error(230, exit_block(Tag)), T, true),
	top_abort(T).


    gui_toplevel1 :-
	repeat,
	    read_command(Command),
	    state_idle(Command),
	fail.


state_idle(call(GoalString)) :- !,
	exec_goal(GoalString).
state_idle(end_of_file) :- !,
	yield(_,_).
state_idle(end) :- !.
%state_idle(more) :- !.
state_idle(exit) :- !,
	exit_block(end).
state_idle(Junk) :- !,
	error(6, state_idle(Junk)).


state_success(State) :-
	read_command(Command),
	state_success(Command,State).

state_success(end_of_file,State) :- !,
	yield(_,_),
	state_success(State).
state_success(more,_) :- !,
	fail.
state_success(end,_) :- !.
state_success(exit,_) :- !,
	exit_block(end).
state_success(Junk, Dgs) :-
	error(6, state_success(Junk, Dgs)).
	

read_goal(GoalString,Goal,Vars) :-
	open(GoalString,string,Stream),
	get_flag(toplevel_module, Module),
	( block(readvar(Stream,Goal,Vars)@Module,_,fail) ->
	    close(Stream)
	;
	    close(Stream),
	    fail
	).

% exec_goal(GoalString)
% read a goal from toplevel_in, run it, do the more-dialog
% and always succeed in the end
exec_goal(GoalString) :-
	read_goal(GoalString,Goal,Vars),	% fails on syntax error
	get_flag(toplevel_module, Module),
	register_inspected_term(Goal, Module),
	printf_with_current_modes_body(answer_out, (?- Goal), Module),
	writeln(answer_out, "."),
	flush(answer_out),
	( get_flag(goal_expansion,on) ->
	    tr_goals(Goal, TransGoal, Module)	% might fail
	;
	    TransGoal=Goal
	),
	!,
	cputime(Tstart),
	(
	    get_cut(Before),
	    top_call_local(TransGoal, Module),
	    Time is cputime-Tstart,
	    get_cut(After),
	    write_vars(answer_out, Vars),
	    write_dgs(answer_out),
	    ( (Before == After ; Vars == []) ->
		Answer = "Yes"
	    ;
		Answer = "More"
	    ),
	    printf(answer_out, "%s (%.2fs cpu)%n%b", [Answer,Time]),
	    write_exdr(toplevel_out, Answer),
	    state_success(0)			% fail if more requested
	->
	    true
	;
	    Time is cputime-Tstart,
	    printf(answer_out, "No (%.2fs cpu)%n%b", [Time]),
	    write_exdr(toplevel_out, "No")
	).
exec_goal(_) :-
	exit_block(abort).


read_command(C) :-
	( select([toplevel_in], 0, []) ->
%	    repeat,
	    flush(output),
	    flush(error),
	    yield(99, _),
	    true		% handle aborts here
%	    ,select([toplevel_in], 0, [_])
	;
	    true
	),
	read_exdr(toplevel_in,C).
%	writeln(got(C)).

drain_stream(S) :-
	( select([S], 0, []) -> true ; get(S, _), drain_stream(S) ).
	

write_vars(Stream, Vars) :-
	get_flag(toplevel_module, Module),
	( foreach([Name|Value],Vars),param(Stream,Module) do
	    write(Stream, Name),
	    write(Stream, ' = '),
	    printf_with_current_modes_body(Stream, Value, Module),
	    nl(Stream)
	).

write_dgs(Stream) :-
	suspensions(Susps),
	length(Susps, N),
	( N > 1 ->
	    printf(Stream, "There are %d delayed goals.%n", N)
	; N > 0 ->
	    printf(Stream, "There is 1 delayed goal.%n", [])
	;
	    true
	).


printf_with_current_modes_body(Stream, Value, Module) :-
	printf_current(Stream, Value, '', Module).

printf_current(Stream, Value, Goal, Module) :-
	get_flag(output_mode,Mode),
	concat_string(['%', Mode, Goal, 'w'], Format),
	printf(Stream, Format, [Value])@Module.

