% ----------------------------------------------------------------------
% System:	ECLiPSe Constraint Logic Programming System
% Copyright (C) Imperial College London and ICL 1995-1999
% Version:	$Id: util.pl,v 1.5 1999/08/07 15:50:47 js10 Exp $
% ----------------------------------------------------------------------

/*
 * SEPIA PROLOG SOURCE MODULE
 *
 *	sccsid("%W%		%E%").
 *	sccscr("%Z%  Copyright 1989 ECRC GmbH ").
 *
 * IDENTIFICATION:	util.pl 
 *
 * DESCRIPTION: 	Various utility predicates, for user convenience.
 *
 * CONTENTS:
 *
 */

:- module_interface(util).
:- begin_module(util).

:- export
	add_path/1,
	add_suffix/1,
	between/3,
	c_compile_and_load/1,
	compiled/0,
	compile_selection/0,
	edit/1,
	file_info/1,
	list_error/3,
	stream/1,
	streams/0,
	read_line/1,
	read_line/2,
	time/1,
	write_history/0.

:- import
	compile_stream_/2,
	current_file/1,
	get_flag_body/4,
	history_collect/1,
	once_body/2
    from sepia_kernel.


% add_path(+Path) - prepend a directory to the library path

add_path(Path) :-
	get_flag(library_path, X),
	set_flag(library_path, [Path|X]).

add_suffix(Suffix) :-
	get_flag(prolog_suffix, X),
	set_flag(prolog_suffix, [Suffix|X]).


% streams - print a table of currently opened streams

streams :-
	current_stream(S),
	get_stream_info(S, device, D),
	get_stream_info(S, mode, M),
	get_stream_info(S, name, N),
	printf("%3d   %6s   %6s   %q%n", [S,D,M,N]),
	fail.
streams.

stream(Stream) :-
	get_stream_info(Stream, _, _),	% so that it fails if not open
	!,
	atom_string('%-20s%w%n', Format),	% to avoid problems in Q mode
	(   get_stream_info(Stream, F, V),
	    printf(Format, [F, V]),
	    fail
	;
	    current_atom(Alias),
	    current_stream(Alias),
	    get_stream(Alias, Stream),
	    printf(Format, [alias, Alias]),
	    fail
	;
	    true
	).

file_info(File) :-
	atom_string('%-20s%w%n', Format),	% to avoid problems in Q mode
	(   get_file_info(File, F, V),
	    printf(Format, [F, V]),
	    fail
	;
	    true
	).

% read_line([+Stream, ] ?String) - read a line of input into String

read_line(Stream, String) :-
	read_string(Stream, end_of_line, _, String).

read_line(String) :-
	read_string(input, end_of_line, _, String).

between(From, To, I) :-
	between(From, To, 1, I).


% time(+Goal) - like call(Goal), but print cputime used

:- tool(time/1, time_body/2).

time_body(Goal, Module) :-
	cputime(T0),
	(
	    call(Goal, Module),
	    true
	->
	    T is cputime - T0,
	    write('\nSuccess, time = '),
	    writeln(T)
	;
	    T is cputime - T0,
	    write('\nFailure, time = '),
	    writeln(T),
	    fail
	).


% Write out the current history.
write_history :-
	history_collect(List),
	open('.eclipse_history', write, Stream),
	write_history(Stream, List).

write_history(Stream, []) :-
	close(Stream).
write_history(Stream, [Goal-_|More]) :-
	printf(Stream, "%OQDw.\n", [Goal]),
	write_history(Stream, More).


% print a list of compiled files and if they were modified since

compiled :-
	current_compiled_file(File, Time, _),
	write(File),
	(get_file_info(File, mtime) =\= Time ->
		writeln(" (modified)")
	;
		nl
	),
	fail.
compiled.

% List all the errors whose message contains a specified text

list_error(String, N, Message) :-
	current_error(N),
	error_id(N, Message),
	substring(Message, String, _).


% Compile selected text (OpenWindow)

:- tool(compile_selection/0, compile_selection/1).

compile_selection(Module) :-
    exec(xv_get_sel, [null, S], Pid),
    compile_stream_(S, Module),
    wait(Pid, _).


% invoke an editor on the source of a predicate

:- tool(edit/1, edit/2).

edit(Pred0, Module0) :-
        ( get_flag_body(Pred0, tool, on, Module0) ->
		tool_body(Pred0, Pred, Module)	% edit the tool body instead
	;
		Pred = Pred0,
		Module = Module0
	),
        get_flag_body(Pred, source_file, File, Module),
        get_flag_body(Pred, source_line, Line, Module),
	( getenv('EDITOR', Editor) -> true ; Editor = "vi"),
        concat_string([Editor, " +", Line, " ", File], Cmd),
	get_file_info(File, mtime, TimeBefore),
        sh(Cmd),
	( get_file_info(File, mtime) =\= TimeBefore ->
                compile(File, Module)		% recompile if changed
        ;
                true
        ).


% Invoke the proper C compiler on File.c and load the result dynamically

c_compile_and_load(File) :-
        get_flag(installation_directory, Inst),
        get_flag(hostarch, Arch),
        get_flag(object_suffix, O),
        concat_string([Inst,"/lib/",Arch,"/Makefile.external"], Makefile),
        concat_string([File,.,O], Ofile),
        concat_string(["sh -c \"ECLIPSEDIR=",Inst,";export ECLIPSEDIR;",
                       "make -f ",Makefile," ",Ofile,"\""], Make),
	writeln(Make),
        exec(Make, []),
	load(Ofile).

