% ----------------------------------------------------------------------
% System:	ECLiPSe Constraint Logic Programming System
% Copyright (C) Imperial College London and ICL 1995-1999
% Version:	$Id: xref.pl,v 1.2 1999/08/07 15:50:48 js10 Exp $
% ----------------------------------------------------------------------

/*
 * SEPIA PROLOG SOURCE MODULE
 */

/*
 * sccsid("%W%		%E%").
 * sccscr("%Z%  Copyright 1989 ECRC GmbH ").
 */

/*
 * IDENTIFICATION:	xref.pl
 *
 * DESCRIPTION: 	Prolog cross-reference tool.
 *
 *
 * CONTENTS:     
 *
 * REVISION HISTORY:
 * AUTHOR	VERSION	DATE	REASON
 * Michel Dorochevsky	2.4	14.3.90
 */


:- module_interface(xref).
:- export xref/1.

:- begin_module(xref).

:- dynamic(def/1).
:- dynamic(undef/1).

% xref/1
% xref(File)
% xref(+)
%
xref(File) :-
	var(File),
	!,
	error(4, xref(File)).
xref([]) :-
	!,
	print_res.
xref([H|T]) :-
	crossref(H),
	xref(T).
xref(File) :-
	((atom(File) ; string(File)) ->
		crossref(File),
		print_res
	;
		error(5, xref(File))
	).

crossref(File) :- 
   write(File),nl,
   open(File,read,Fin),
   repeat,
   read(Fin, Clause),
   ( Clause = end_of_file, !;
     crossref0(Clause),fail ),
   close(Fin).



% print_res/0
%
print_res :-
   F = output,
   write(F, "Undefined:\n"),
   write_result(F, undef),
   write(F, "Defined:\n"),
   write_result(F, def).


% write_result/1
% write_result(F, Pattern)
% write_result(+,+)
%
write_result(F, Pattern) :-
   get_list(Pattern,List),
   sort(List,SortedList),
   write_list(F, SortedList).



% get_list/2
% get_list(Pattern,List)
% get_list(+,+)
%
get_list(Pattern,List) :-
	Term =.. [Pattern,Pred],
	retract(Term),
	!,
	List = [Pred|RestList],
	get_list(Pattern,RestList).
get_list(_,[]).



% write_list/2
% write_list(OutFile, List)
% write_list(+,+)
%
write_list(F, []) :-
   nl(F).
write_list(F, [Pred|Preds]) :-
   write("   "), writeln(F, Pred), 
   write_list(F, Preds).




% crossref0/1
% crossref0(Clause)
% crossref0(+)
%
crossref0('?-'(Head,Tail)) :-
	!,
	treat_head(Head),
	treat_tail(Tail).
crossref0(':-'(Head,Tail)) :-
	!,
	treat_head(Head),
	treat_tail(Tail).
crossref0(':-'(Tail)) :-
	!,
	check_op_directives(Tail),
	treat_tail(Tail).
crossref0('?-'(Tail)) :-
	!,
	check_op_directives(Tail),
	treat_tail(Tail).
crossref0(Head) :-
	!,
	treat_head(Head).
   


% check_op_directives/1
% check_op_directives(Tail),
% check_op_directives(?),
%
check_op_directives(Directive) :-
	nonvar(Directive),
	functor(Directive,op,3),
	!,
	Directive.
check_op_directives(_).



% treat_head/1
% treat_head(Head)
% treat_head(0)
%
treat_head(Head) :-
   functor(Head, Fun, Ar),
   treat_head_elt(Fun/Ar).



% treat_head_elt/1
% treat_head_elt(Functor)
% treat_head_elt(+)
%
treat_head_elt(X) :-
   def(X),		% already defined
   !.
treat_head_elt(X) :-
   undef(X),		% undefined so far
   !,
   assert(def(X)),
   retract(undef(X)). 
treat_head_elt(X) :-
   assert(def(X)).	% neither defined nor undefined: first occurence



% treat_tail/1
% treat_tail(Goals)
% treat_tail(+)
%
treat_tail(X) :-
   var(X),
   !,
   treat_tail(call(X)).
treat_tail(';'(X,Y)) :-
   !,
   treat_tail(X),
   treat_tail(Y).
treat_tail(','(X,Y)) :-
   !,
   treat_tail(X),
   treat_tail(Y).
treat_tail('|'(X,Y)) :-
   !,
   treat_tail(X),
   treat_tail(Y).
treat_tail('->'(X,Y)) :-
   !,
   treat_tail(X),
   treat_tail(Y).
treat_tail((X '->' Y ';' Z)) :-
   !,
   treat_tail(X),
   treat_tail(Y),
   treat_tail(Z).
treat_tail(not(X)) :-
   !,
   treat_tail(X).
treat_tail('\\+'(X)) :-
   !,
   treat_tail(X).
treat_tail(X) :-
   functor(X, Fun, Ar),
   treat_tail_elt(Fun/Ar).



% treat_tail_elt/1
% treat_tail_elt(Functor)
% treat_tail_elt(+)
%
treat_tail_elt(!/0) :- !.
treat_tail_elt(fail/0) :- !.
treat_tail_elt(true/0) :- !.
treat_tail_elt(X) :-
   def(X), !.
treat_tail_elt(X) :-
   undef(X), !.
treat_tail_elt(X) :-
   assert(undef(X)).
